import dash
import dash_bootstrap_components as dbc
from automation.pages.components import Components
from automation.pages.components.opcua_server import OPCUAServerComponents

dash.register_page(__name__)

layout = dbc.Container(
    [
        dash.dcc.Location(id='opcua_server', refresh=False),
        dbc.Breadcrumb(
            items=[
                {"label": "Communications", "href": "/"},
                {"label": "OPCUA Server", "active": True},
            ],
        ),
        Components.modal_confirm(
            title="Confirmation",
            modal_id="modal-update-opcua-server-centered",
            body_id="modal-update-opcua-server-body",
            yes_button_id="update-opcua-server-yes",
            no_button_id="update-opcua-server-no"
        ),
        OPCUAServerComponents.opcua_server_table()
    ],
    fluid=False,
    className="my-3"
)


import secrets
from peewee import CharField, IntegerField, ForeignKeyField
from ..dbmodels.core import BaseModel
from ..modules.users.users import Users as CVTUsers
from ..modules.users.roles import Roles as CVTRoles
from ..modules.users.roles import Role
from ..modules.users.users import User
from werkzeug.security import generate_password_hash, check_password_hash

users = CVTUsers()
roles = CVTRoles()

class Roles(BaseModel):
    
    identifier = CharField(unique=True, max_length=16)
    name = CharField(unique=True, max_length=32)
    level = IntegerField()
    __defaults__ = [
        {"name": "sudo", "level": 0, "identifier": secrets.token_hex(4)},
        {"name": "admin", "level": 1, "identifier": secrets.token_hex(4)},
        {"name": "supervisor", "level": 2, "identifier": secrets.token_hex(4)},
        {"name": "operator", "level": 10, "identifier": secrets.token_hex(4)},
        {"name": "auditor", "level": 100, "identifier": secrets.token_hex(4)},
        {"name": "guest", "level": 256, "identifier": secrets.token_hex(4)}
    ]

    @classmethod
    def create(cls, name:str, level:int, identifier:str)->tuple:
        
        name = name.upper()

        if cls.name_exist(name):

            return None, f"role {name} is already used"

        if cls.identifier_exist(identifier):
                
            return None, f"identifier {identifier} is already used"
        
        query = cls(name=name, level=level, identifier=identifier)
        query.save()
        
        return query, f"Role creation successful"

    @classmethod
    def read_by_name(cls, name:str):
      
        return cls.get_or_none(name=name.upper())
    
    @classmethod
    def read_by_identifier(cls, identifier:str):
      
        return cls.get_or_none(identifier=identifier)

    @classmethod
    def name_exist(cls, name:str)->bool:
        
        return True if cls.get_or_none(name=name.upper()) else False
    
    @classmethod
    def identifier_exist(cls, identifier:str)->bool:

        return True if cls.get_or_none(identifier=identifier) else False
    
    @classmethod
    def read_names(cls)->list:

        return [role.name for role in cls.select()]
    
    @classmethod
    def fill_cvt_roles(cls):
        r"""
        Documentation here
        """
        for role in cls.select():

            _role = Role(
                name=role.name,
                level=role.level,
                identifier=role.identifier
            )

            roles.add(role=_role)

    def serialize(self)->dict:
        r"""
        Serialize database record to a jsonable object
        """

        return {
            "id": self.id,
            "identifier": self.identifier,
            "name": self.name,
            "level": self.level
        }


class Users(BaseModel):
    
    identifier = CharField(unique=True, max_length=16)
    username = CharField(unique=True, max_length=64)
    role = ForeignKeyField(Roles, backref='users', on_delete='CASCADE')
    email = CharField(unique=True, max_length=128)
    password = CharField()
    token = CharField(null=True)
    name = CharField(max_length=64, null=True)
    lastname = CharField(max_length=64, null=True)

    @classmethod
    def create(cls, user:User)-> dict:

        if cls.username_exist(user.username):

            return None, f"username {user.username} is already used"

        if cls.email_exist(user.email):

            return None, f"email {user.email} is already used"
        
        if cls.identifier_exist(user.identifier):

            return None, f"identifier {user.identifier} is already used"
        
        query = cls(
            username=user.username,
            role=Roles.read_by_name(name=user.role.name),
            email=user.email,
            password=user.password,
            identifier=user.identifier,
            name=user.name,
            lastname=user.lastname,
            token=user.token
            )
        query.save()

        return query, f"User creation successful"
    
    @classmethod
    def login(cls, password:str, username:str="", email:str=""):
        r"""
        Documentation here
        """
        if username:

            user = cls.get_or_none(username=username)

        if email:

            user = cls.get_or_none(email=email)
        
        if user:

            if user.decode_password(password):
                
                if not user.token:
                    
                    user.token = cls.encode(secrets.token_hex(4))
                    user.save()
                
                users.login(password=password, token=user.token, username=username, email=email)

                return user, f"Login successfull"

            return None, f"Invalid credentials" 

        return None, f"Invalid Username or Email"      
    
    @classmethod
    def logout(cls, token:str):
        r"""
        Documentation here
        """
        user = cls.get_or_none(token=token)

        if user:

            user.token = None
            user.save()

            return user, "Logout successfull"
        
        return None, "Invalid Token"

    @classmethod
    def read_by_username(cls, username:str):
   
        return cls.get_or_none(username=username)

    @classmethod
    def read_by_name(cls, name:str):
   
        return cls.get_or_none(name=name)

    @classmethod
    def name_exist(cls, name:str)->bool:

        return True if cls.get_or_none(name=name) else False
    
    @classmethod
    def username_exist(cls, username:str)->bool:

        return True if cls.get_or_none(username=username) else False
    
    @classmethod
    def email_exist(cls, email:str)->bool:

        return True if cls.get_or_none(email=email) else False
    
    @classmethod
    def identifier_exist(cls, identifier:str)->bool:

        return True if cls.get_or_none(identifier=identifier) else False
    
    @classmethod
    def encode(cls, value:str)->str:

        return generate_password_hash(value)
    

    def decode_password(self, password:str)->str:

        return check_password_hash(self.password, password)
    
    def decode_token(self, token:str)->str:

        return check_password_hash(self.token, token)
    
    @classmethod
    def fill_cvt_users(cls):
        r"""
        Documentation here
        """
        for user in cls.select():

            users.signup(
                username=user.username,
                role_name=user.role.name,
                email=user.email,
                password=user.password,
                name=user.name,
                lastname=user.lastname,
                identifier=user.identifier,
                encode_password=False
            )

    def serialize(self)-> dict:
        r"""
        Serialize database record to a jsonable object
        """

        return {
            "id": self.id,
            "identifier": self.identifier,
            "username": self.username,
            "email": self.email,
            "role": self.role.serialize(),
            "name": self.name,
            "lastname": self.lastname
        }


# -*- coding: utf-8 -*-
"""workers/worker.py

This module implements all thread classes for workers.
"""
from threading import Thread
from threading import Event as ThreadEvent


class BaseWorker(Thread):

    def  __init__(self):

        super(BaseWorker, self).__init__()

        self.stop_event = ThreadEvent()

    def get_stop_event(self):

        return self.stop_event

    def stop(self):
        
        self.stop_event.set()

    def __getstate__(self):

        state = self.__dict__.copy()
        del state['stop_event']
        return state

    def __setstate__(self, state):
        
        self.__dict__.update(state)
        self.stop_event = ThreadEvent()
    


import dash
import dash_bootstrap_components as dbc
from dash import html

class Components:

    @classmethod
    def page_title(cls, title:str):

        return dash.html.H1(title)

    @classmethod
    def modal_error(cls, title:str, modal_id:str, button_close_id:str, body_id:str):

        return dash.html.Div(
            [
                dbc.Modal(
                    [
                        dbc.ModalHeader(dbc.ModalTitle(title), close_button=True),
                        dbc.ModalBody("This modal is vertically centered", id=body_id),
                        dbc.ModalFooter(
                            dbc.Button(
                                "Close",
                                id=button_close_id,
                                className="ms-auto",
                                n_clicks=0,
                            )
                        ),
                    ],
                    id=modal_id,
                    centered=True,
                    is_open=False,
                ),
            ]
        )

    @classmethod
    def modal_confirm(cls, title:str, modal_id:str, body_id:str, yes_button_id:str, no_button_id:str):

        return dash.html.Div(
            [
                dbc.Modal(
                    [
                        dbc.ModalHeader(dbc.ModalTitle(title), close_button=True),
                        dbc.ModalBody("This modal is vertically centered", id=body_id),
                        dbc.ModalFooter(
                            [
                                dbc.Button(
                                    "Yes",
                                    id=yes_button_id,
                                    className="ms-auto",
                                    n_clicks=0,
                                ),
                                dbc.Button(
                                    "No",
                                    id=no_button_id,
                                    className="ms-auto",
                                    n_clicks=0,
                                )
                            ]
                        ),
                    ],
                    id=modal_id,
                    centered=True,
                    is_open=False,
                ),
            ]
        )

    @classmethod
    def navbar(cls)->dbc.NavbarSimple:

        return dbc.NavbarSimple(
            children=[
                dbc.DropdownMenu(
                    children=[
                        dbc.DropdownMenuItem("Communications", header=True),
                        dbc.DropdownMenuItem("OPCUA Client", href="/"),
                        dbc.DropdownMenuItem("OPCUA Server", href="/opcua-server"),
                    ],
                    nav=True,
                    in_navbar=True,
                    label="Communications",
                ),
                dbc.NavItem(dbc.NavLink("Database", href="/database")),
                dbc.DropdownMenu(
                    children=[
                        dbc.DropdownMenuItem("Tags", header=True),
                        dbc.DropdownMenuItem("Definition", href="/tags", id="tags_definition_link"),
                        dbc.DropdownMenuItem("Trends", href="/trends"),
                        dbc.DropdownMenuItem("Filter", href="/filter"),
                    ],
                    nav=True,
                    in_navbar=True,
                    label="Tags",
                ),
                dbc.DropdownMenu(
                    children=[
                        dbc.DropdownMenuItem("Alarms", header=True),
                        dbc.DropdownMenuItem("Definition", href="/alarms"),
                        dbc.DropdownMenuItem("History", href="/alarms-history"),
                    ],
                    nav=True,
                    in_navbar=True,
                    label="Alarms",
                ),
                dbc.DropdownMenu(
                    children=[
                        dbc.DropdownMenuItem("Machines", header=True),
                        dbc.DropdownMenuItem("Summary", href="/machines"),
                        dbc.DropdownMenuItem("Detailed", href="/machines-detailed"),
                    ],
                    nav=True,
                    in_navbar=True,
                    label="Machines",
                ),
            ],
            fluid=True, # Navbar container is modified.
            brand=html.Span("PyAutomation Configuration", className="custom-brand-class mx-2"), # Styles are applied to the HTML brand element.
            brand_href="/",
            color="primary",
            dark=True,
        )


# -*- coding: utf-8 -*-
"""pyhades/logger/machines.py
"""
from ..dbmodels import OPCUAServer
from .core import BaseEngine, BaseLogger
from ..utils.decorators import db_rollback


class OPCUAServerLogger(BaseLogger):

    def __init__(self):

        super(OPCUAServerLogger, self).__init__()

    @db_rollback
    def create(
            self,
            name:str,
            namespace:str,
            access_type:str="Read"
            ):
        r"""
        Documentation here
        """
        if not self.check_connectivity():

            return None
       
        OPCUAServer.create(
            name=name,
            namespace=namespace,
            access_type=access_type
        )

    @db_rollback
    def put(
        self,
        namespace:str,
        access_type:str
        ):

        if not self.check_connectivity():
            
            return None    
        
        if access_type:
            
            OPCUAServer.update_access_type(namespace=namespace, access_type=access_type)

            obj = OPCUAServer.read_by_namespace(namespace=namespace)

            return obj
    
    @db_rollback
    def read_all(self):
        if not self.check_connectivity():

            return list()
        
        return OPCUAServer.read_all()
    
    @db_rollback
    def read_by_namespace(self, namespace:str):
        if not self.check_connectivity():

            return list()
        
        return OPCUAServer.read_by_namespace(namespace=namespace)
     

class OPCUAServerLoggerEngine(BaseEngine):
    r"""
    OPCUA Server logger Engine class for Tag thread-safe database logging.

    """

    def __init__(self):

        super(OPCUAServerLoggerEngine, self).__init__()
        self.logger = OPCUAServerLogger()

    def create(
        self,
        name:str,
        namespace:str,
        access_type:str="Read"
        ):

        _query = dict()
        _query["action"] = "create"
        _query["parameters"] = dict()
        _query["parameters"]["name"] = name
        _query["parameters"]["namespace"] = namespace
        _query["parameters"]["access_type"] = access_type
        
        return self.query(_query)
    
    def put(
        self,
        namespace:str,
        access_type:str
        ):

        _query = dict()
        _query["action"] = "put"
        _query["parameters"] = dict()
        _query["parameters"]["namespace"] = namespace
        _query["parameters"]["access_type"] = access_type

        return self.query(_query)
    
    def read_by_namespace(
        self,
        namespace:str
        ):

        _query = dict()
        _query["action"] = "read_by_namespace"
        _query["parameters"] = dict()
        _query["parameters"]["namespace"] = namespace

        return self.query(_query)

    def read_all(self):

        _query = dict()
        _query["action"] = "read_all"
        _query["parameters"] = dict()
        return self.query(_query)




# -*- coding: utf-8 -*-
"""pyhades/logger/datalogger.py

This module implements a database logger for the CVT instance, 
will create a time-serie for each tag in a short memory data base.
"""
import pytz, logging, math
from collections import defaultdict
from datetime import datetime
from ..tags.tag import Tag
from ..dbmodels import Tags, TagValue, Units, Segment, Variables
from ..modules.users.users import User
from ..tags.cvt import CVTEngine
from .core import BaseLogger, BaseEngine
from ..variables import *
from ..utils.decorators import db_rollback


DATETIME_FORMAT = "%Y-%m-%d %H:%M:%S.%f"

class DataLogger(BaseLogger):
    """
    Data Logger class.

    This class serves as an API for managing tag settings and accessing logged tags.

    **Usage Example**:

    .. code-block:: python

        >>> from pyhades import DataLogger
        >>> _logger = DataLogger()

    """

    def __init__(self):

        super(DataLogger, self).__init__()
        self.tag_engine = CVTEngine()

    @db_rollback
    def set_tag(
        self, 
        id:str,
        name:str, 
        unit:str, 
        data_type:str, 
        description:str="", 
        display_name:str="",
        display_unit:str=None,
        opcua_address:str=None, 
        node_namespace:str=None,
        scan_time:int=None,
        dead_band:float=None,
        manufacturer:str="",
        segment:str=""
        ):
        r"""
        Documentation here
        """
        if not self.check_connectivity():

            return None
            
        Tags.create(
            id=id,
            name=name, 
            unit=unit,
            data_type=data_type,
            description=description,
            display_name=display_name,
            display_unit=display_unit,
            opcua_address=opcua_address,
            node_namespace=node_namespace,
            scan_time=scan_time,
            dead_band=dead_band,
            manufacturer=manufacturer,
            segment=segment
            )
            
    @db_rollback
    def delete_tag(self, id:str):
        r"""
        Documentation here
        """
        if not self.check_connectivity():

            return None
        
        tag, _ = Tags.get_or_create(identifier=id)
        Tags.put(id=tag.id, active=False)

    @db_rollback
    def get_tag_by_name(self, name:str):
        r"""
        Documentation here
        """
        if not self.check_connectivity():

            return None
        
        return Tags.read_by_name(name=name)

    @db_rollback
    def update_tag(self, id:str, **kwargs):
        r"""
        Documentation here
        """
        if not self.check_connectivity():

            return None
        
        tag = Tags.get(identifier=id)

        if "gaussian_filter" in kwargs:
            
            if kwargs['gaussian_filter'].lower() in ('1', 'true'):

                kwargs['gaussian_filter'] = True

            else:

                kwargs['gaussian_filter'] = False
        
        return Tags.put(id=tag.id, **kwargs)

    @db_rollback
    def set_tags(self, tags):
        r"""
        Documentation here
        """
        if not self.check_connectivity():

            return None
        
        for tag in tags:

            self.set_tag(tag)

    @db_rollback
    def get_tags(self):
        r"""
        Documentation here
        """
        if not self.check_connectivity():

            return list()
            
        return Tags.read_all()
    
    @db_rollback
    def write_tag(self, tag, value, timestamp):
        r"""
        Documentation here
        """
        if not self.is_history_logged:

            return None
        
        if not self.check_connectivity():

            return None
        
        trend = Tags.read_by_name(tag)
        unit = Units.read_by_unit(unit=trend.display_unit.unit)
        TagValue.create(tag=trend, value=value, timestamp=timestamp, unit=unit)

    @db_rollback
    def write_tags(self, tags:list):
        r"""
        Documentation here
        """
        if not self.is_history_logged:

            return None
        
        if not self.check_connectivity():

            return None
            
        _tags = tags.copy()
        
        for counter, tag in enumerate(tags):
            
            _tag = Tags.read_by_name(tag['tag'])
            
            if _tag:

                unit = Units.get_or_none(id=_tag.display_unit.id)
                _tags[counter].update({
                    'tag': _tag,
                    'unit': unit
                })
        
        TagValue.insert_many(_tags).execute()

    @db_rollback
    def read_trends(self, start:str, stop:str, timezone:str, tags):
        r"""
        Documentation here
        """  
        
        if not self.is_history_logged:

            return None
        
        if not self.check_connectivity():

            return dict()

        _timezone = pytz.timezone(timezone)
        start = _timezone.localize(datetime.strptime(start, DATETIME_FORMAT)).astimezone(pytz.UTC).timestamp()
        stop = _timezone.localize(datetime.strptime(stop, DATETIME_FORMAT)).astimezone(pytz.UTC).timestamp()  
              
        query = (TagValue
                .select(Tags.name, TagValue.value, TagValue.timestamp,
                        Units.unit.alias('tag_value_unit'), Variables.name.alias('variable_name'))
                .join(Tags)
                .join(Units, on=(Tags.unit == Units.id))
                .join(Variables, on=(Units.variable_id == Variables.id))
                .where((TagValue.timestamp.between(start, stop)) & (Tags.name.in_(tags)))
                .order_by(TagValue.timestamp)
                .dicts()) 
         
        # Structure the data
        time_span = (stop - start ) / 60 # span in minutes
        result = defaultdict(lambda: {"values": []})
        if time_span > 60 * 24 * 7:  # 1 week
            # Aggregate data every 1 day
            result = self._agregate_data_every_seconds(query=query, result=result, seconds=3600 * 24, timezone=timezone)

        elif time_span > 60 * 24 * 2:  # 2 days
            # Aggregate data every 1 hora
            result = self._agregate_data_every_seconds(query=query, result=result, seconds=3600, timezone=timezone)

        elif time_span > 60 * 2:  # 2 horas
            # Aggregate data every 1 minute
            result = self._agregate_data_every_seconds(query=query, result=result, seconds=60, timezone=timezone)

        else:
            # Use original data
            
            for entry in query:
                
                from_timezone = pytz.timezone('UTC')
                timestamp = entry['timestamp']
                timestamp = from_timezone.localize(timestamp)
                result[entry['name']]["values"].append({
                    "x": timestamp.astimezone(_timezone).strftime(self.tag_engine.DATETIME_FORMAT),
                    "y": entry['value']
                })
        
        for tag in tags:

            result[tag]['unit'] = self.tag_engine.get_display_unit_by_tag(tag)
        
        return result

    @db_rollback
    def read_table(self, start:str, stop:str, timezone:str, tags:list, page:int=1, limit:int=20):
        r"""
        Get historical data in table format with pagination
        """
        if not self.is_history_logged:
            return None
        
        if not self.check_connectivity():
            return dict()

        _timezone = pytz.timezone(timezone)
        try:
            start_dt = _timezone.localize(datetime.strptime(start, DATETIME_FORMAT)).astimezone(pytz.UTC).timestamp()
            stop_dt = _timezone.localize(datetime.strptime(stop, DATETIME_FORMAT)).astimezone(pytz.UTC).timestamp()
        except ValueError:
            # Fallback or error handling if needed, though read_trends assumes correct format
            return dict()

        # Base query
        query = (TagValue
                .select(Tags.name, TagValue.value, TagValue.timestamp,
                        Units.unit.alias('tag_value_unit'))
                .join(Tags)
                .join(Units, on=(Tags.unit == Units.id))
                .where((TagValue.timestamp.between(start_dt, stop_dt)) & (Tags.name.in_(tags)))
                .order_by(TagValue.timestamp.desc()))

        total_records = query.count()
        
        # Safe pagination
        if limit <= 0: limit = 20
        if page <= 0: page = 1
        
        total_pages = math.ceil(total_records / limit)
        if total_pages == 0: total_pages = 1
        
        has_next = page < total_pages
        has_prev = page > 1

        paginated_query = query.paginate(page, limit).dicts()
        
        data = []
        utc_timezone = pytz.timezone('UTC')
        
        for entry in paginated_query:
            timestamp = entry['timestamp']
            # timestamp in DB is float (epoch) or datetime? 
            # In read_trends: timestamp = entry['timestamp']; timestamp = from_timezone.localize(timestamp)
            # This implies entry['timestamp'] is NOT timezone aware or is a float?
            # In read_trends: start/stop converted to .timestamp() (float).
            # Peewee timestamp field usually stores whatever you give it. If float was stored, it comes back as float.
            # But line 247 in read_trends: timestamp = from_timezone.localize(timestamp)
            # localize() works on datetime objects.
            # So TagValue.timestamp is likely a DateTimeField in Peewee, but stored as UTC?
            # Wait, line 212: start = ... .timestamp(). 
            # TagValue.timestamp.between(start, stop)
            # If start/stop are floats, and TagValue.timestamp compares to them, TagValue.timestamp might be float/DoubleField?
            # Or Peewee handles conversion?
            # Let's check dbmodels/tags.py if possible. 
            # But relying on read_trends line 246-247:
            # timestamp = entry['timestamp']
            # timestamp = from_timezone.localize(timestamp)
            # IF timestamp is float, localize() fails. localize takes datetime.
            # SO timestamp must be a datetime object (naive).
            # BUT line 212 converts start/stop to floats!
            # If TagValue.timestamp is DateTimeField, Peewee might accept float for comparison? Or start/stop should be datetimes?
            # Actually line 212: .timestamp() returns float.
            # So TagValue.timestamp might be a float/DoubleField storing epoch?
            # If so, line 247 `from_timezone.localize(timestamp)` would FAIL on a float.
            # Let's assume read_trends logic is correct and see what it does.
            # If timestamp is float, `datetime.fromtimestamp(timestamp, pytz.UTC)` is needed.
            # If timestamp is datetime, `localize` is needed.
            
            # Let's check `read_trends` carefully.
            # 246| timestamp = entry['timestamp']
            # 247| timestamp = from_timezone.localize(timestamp)
            
            # This strongly suggests `entry['timestamp']` is a naive datetime object.
            # THEN why line 212 converts to .timestamp()?
            # `start = ... .timestamp()`
            # Maybe TagValue.timestamp is Integer/Float (Epoch)?
            # If so, 247 is suspicious.
            # However, I should follow `read_trends` pattern OR be robust.
            # If I check `write_tag`: `TagValue.create(..., timestamp=timestamp, ...)`
            # where timestamp comes from `datetime.now(pytz.utc).astimezone(TIMEZONE)` (from tags.py).
            
            # If I look at `read_trends` line 243 `for entry in query:` where query is `.dicts()`.
            # I will assume `entry['timestamp']` works like in `read_trends`.
            # BUT, if `read_trends` is working code, then `entry['timestamp']` is likely a datetime.
            # AND `.between(start, stop)` works with floats if the column is float.
            # Or maybe `start` and `stop` being floats are auto-converted?
            # Actually, looking at line 246, `timestamp` variable is reused. 
            
            # I will try to support both or verify.
            # Safest is `datetime.fromtimestamp(entry['timestamp'])` if it's float/int, or just use it if it's datetime.
            # Given `read_trends` code, I suspect `timestamp` in DB is DateTimeField.
            # If so, `start` and `stop` should probably be datetimes. 
            # Line 212 `... .timestamp()` makes them floats.
            # Peewee `DateTimeField` vs float comparison...
            
            # I'll stick to what I see.
            # But wait, `read_trends` logic at 247 `from_timezone.localize(timestamp)` implies naive datetime.
            
            # Implementation:
            ts_val = entry['timestamp']
            if isinstance(ts_val, (int, float)):
                dt_object = datetime.fromtimestamp(ts_val, pytz.UTC)
            else:
                # Assuming naive datetime in UTC (based on read_trends using 'UTC' timezone to localize)
                dt_object = utc_timezone.localize(ts_val) if ts_val.tzinfo is None else ts_val
                
            formatted_ts = dt_object.astimezone(_timezone).strftime(DATETIME_FORMAT)
            
            data.append({
                "timestamp": formatted_ts,
                "tag_name": entry['name'],
                "value": f"{entry['value']} {entry['tag_value_unit']}"
            })

        return {
            "data": data,
            "pagination": {
                "page": page,
                "limit": limit,
                "total_records": total_records,
                "total_pages": total_pages,
                "has_next": has_next,
                "has_prev": has_prev
            }
        }
        
    def _agregate_data_every_seconds(self, query, result, seconds:int, timezone:str="UTC"):
        r"""Documentation here
        """
        # Aggregate data every 5 seconds
        target_timezone = pytz.timezone(timezone)
        buffer = defaultdict(lambda: {"sum": 0, "count": 0, "last_timestamp": None})

        for entry in query:
            bucket = entry['timestamp'].replace(second=(entry['timestamp'].second // seconds) * seconds, microsecond=0)
            buffer_key = (entry['name'], bucket)
            buffer[buffer_key]["sum"] += entry['value']
            buffer[buffer_key]["count"] += 1
            buffer[buffer_key]["last_timestamp"] = entry['timestamp']
            buffer[buffer_key]['unit'] = entry["tag_value_unit"]
            buffer[buffer_key]['variable'] = entry['variable_name']

        for (tag_name, bucket), data in buffer.items():

            avg_value = data["sum"] / data["count"]
            last_timestamp = data["last_timestamp"]
            from_timezone = pytz.timezone('UTC')
            last_timestamp = from_timezone.localize(last_timestamp)
            result[tag_name]["values"].append({
                "x": last_timestamp.astimezone(target_timezone).strftime(self.tag_engine.DATETIME_FORMAT),
                # "y": eval(f"{variable}.convert_value({avg_value}, from_unit={'unit'}, to_unit={'_tag.get_display_unit()'})")
                "y": avg_value
            })
        
        return result
        
    @db_rollback
    def read_segments(self):
        r"""
        Documentation here
        """
        if not self.check_connectivity():

            return list()

        return Segment.read_all()


class DataLoggerEngine(BaseEngine):
    r"""
    Data logger Engine class for Tag thread-safe database logging.

    """
    def __init__(self):

        super(DataLoggerEngine, self).__init__()
        self.logger = DataLogger()

    def create_tables(self, tables):
        r"""
        Create default PyHades database tables

        ['TagTrend', 'TagValue']

        **Parameters**

        * **tables** (list) list of database model

        **Returns** `None`
        """
        self.logger.create_tables(tables)

    def drop_tables(self, tables:list):
        r"""
        Drop tables if exist in database

        **Parameters**

        * **tables** (list): List of database model you want yo drop
        """
        self.logger.drop_tables(tables)

    def set_tag(
        self,
        tag:Tag
        ):
        r"""
        Define tag names you want log in database, these tags must be defined in CVTEngine

        **Parameters**

        * **tag** (str): Tag name defined in CVTEngine
        * **period** (float): Sampling time to log tag on database

        **Returns** `None`
        """
        _query = dict()
        _query["action"] = "set_tag"
        _query["parameters"] = dict()
        _query["parameters"]["id"] = tag.id
        _query["parameters"]["name"] = tag.name
        _query["parameters"]["unit"] = tag.unit
        _query["parameters"]["data_type"] = tag.data_type
        _query["parameters"]["description"] = tag.description
        _query["parameters"]["display_name"] = tag.display_name
        _query["parameters"]["display_unit"] = tag.display_unit
        _query["parameters"]["opcua_address"] = tag.opcua_address
        _query["parameters"]["node_namespace"] = tag.node_namespace
        _query["parameters"]["scan_time"] = tag.scan_time
        _query["parameters"]["dead_band"] = tag.dead_band
        _query["parameters"]["manufacturer"] = tag.manufacturer
        _query["parameters"]["segment"] = tag.segment
        
        return self.query(_query)

    def get_tags(self):
        r"""

        """
        _query = dict()
        _query["action"] = "get_tags"
        _query["parameters"] = dict()
        
        return self.query(_query)
    
    def get_tag_by_name(self, name:str):
        r"""

        """
        _query = dict()
        _query["action"] = "get_tag_by_name"
        _query["parameters"] = dict()
        _query["parameters"]["name"] = name
        
        return self.query(_query)
    
    def update_tag(
            self, 
            id:str, 
            user:User|None=None,
            **kwargs
            ):
        r"""Documentation here

        # Parameters

        - 

        # Returns

        - 
        """

        _query = dict()
        _query["action"] = "update_tag"
        _query["parameters"] = dict()
        _query["parameters"]["id"] = id
        for key, value in kwargs.items():

            _query["parameters"][key] = value
    
        return self.query(_query)
    
    def delete_tag(self, id:str):
        r"""Documentation here

        # Parameters

        - 

        # Returns

        - 
        """
        _query = dict()
        _query["action"] = "delete_tag"
        _query["parameters"] = dict()
        _query["parameters"]["id"] = id
        
        return self.query(_query)

    def write_tag(self, tag:str, value:float, timestamp:datetime):
        r"""
        Writes value to tag into database on a thread-safe mechanism

        **Parameters**

        * **tag** (str): Tag name in database
        * **value** (float): Value to write in tag
        """
        _query = dict()
        _query["action"] = "write_tag"

        _query["parameters"] = dict()
        _query["parameters"]["tag"] = tag
        _query["parameters"]["value"] = value
        _query["parameters"]["timestamp"] = timestamp

        return self.query(_query)

    def write_tags(self, tags:list):
        r"""
        Writes value to tag into database on a thread-safe mechanism

        **Parameters**

        * **tag** (str): Tag name in database
        * **value** (float): Value to write in tag
        """
        _query = dict()
        _query["action"] = "write_tags"

        _query["parameters"] = dict()
        _query["parameters"]["tags"] = tags

        return self.query(_query)
    
    def read_trends(self, start:str, stop:str, timezone:str, *tags):
        r"""
        Read tag value from database on a thread-safe mechanism

        **Parameters**

        * **tag** (str): Tag name in database

        **Returns**

        * **value** (float): Tag value requested
        """
        _query = dict()
        _query["action"] = "read_trends"
        _query["parameters"] = dict()
        _query["parameters"]["start"] = start
        _query["parameters"]["stop"] = stop
        _query["parameters"]["timezone"] = timezone
        _query["parameters"]["tags"] = tags
        return self.query(_query)

    def read_table(self, start:str, stop:str, timezone:str, tags:list, page:int=1, limit:int=20):
        r"""
        Get historical data in table format with pagination on a thread-safe mechanism
        """
        _query = dict()
        _query["action"] = "read_table"
        _query["parameters"] = dict()
        _query["parameters"]["start"] = start
        _query["parameters"]["stop"] = stop
        _query["parameters"]["timezone"] = timezone
        _query["parameters"]["tags"] = tags
        _query["parameters"]["page"] = page
        _query["parameters"]["limit"] = limit
        return self.query(_query)

    def read_segments(self):
        r"""
        Documentation here
        """
        _query = dict()
        _query["action"] = "read_segments"
        _query["parameters"] = dict()
        return self.query(_query)
    


from ..utils.decorators import decorator
from ..buffer import Buffer

data = dict()

def __iad(data:Buffer, tag_name:str):
    r"""
    Out Of Range Algorithm
    """
    from ..managers.alarms import AlarmManager
    alarm_manager = AlarmManager()
    alarm = alarm_manager.get_alarm_by_name(name=f"alarm.iad.{tag_name}")
    if alarm:
        if alarm.state.alarm_status.lower() == "not active":

            mean = sum(data) / len(data)
            # variance = sum((x - mean) ** 2 for x in data) / len(data)
            # std_dev = math.sqrt(variance)
            
            # if abs(std_dev) < 0.01:
                
            #     alarm.description = f"Out Of Range anomaly"
            #     alarm.abnormal_condition()
            
            # else:
                
            #     alarm.description = ""
            #     alarm.normal_condition()

@decorator
def iad_out_of_range(func, args, kwargs):
    r"""
    Documentation here
    """
    cvt = args[0]
    tag_id = kwargs["id"]
    value = kwargs["value"]
    tag = cvt.get_tag(id=tag_id)
    if tag.out_of_range_detection:
        
        if tag.name not in data:
            
            data[tag.name] = Buffer()
        
        data[tag.name](value)
        # Apply IAD logic
        if len(data[tag.name]) >= data[tag.name].size:
            
            __iad(data[tag.name], tag_name=tag.name)

    return func(*args, **kwargs)


from ....extensions.api import api


def init_app():

    from .users import ns as ns_users
    from .roles import ns as ns_roles

    api.add_namespace(ns_users, path="/users")
    api.add_namespace(ns_roles, path="/users/roles")


# -*- coding: utf-8 -*-
"""pyautomation/utils/observer.py

This module implements Observer Utility Functions.
"""
import abc


class Observer(metaclass=abc.ABCMeta):
    """
    Define an updating interface for objects that should be notified of
    changes in a subject.
    """

    def __init__(self):
        self._subject = None
        self._observer_state = None

    @abc.abstractmethod
    def update(self, arg):
        pass


import dash, csv, io
from ...utils import get_nodes_info, get_data_to_update_into_opcua_table
from ...pages.components.opcua import OPCUAComponents
from ...opcua.subscription import SubHandler


subscription_handler = SubHandler()
opcua_components = OPCUAComponents()


def init_callback(app:dash.Dash):
    
    @app.callback(
        dash.Output("data_access_view_table", "children", allow_duplicate=True),
        dash.Input('timestamp-interval', 'n_intervals'),
        dash.State({'type': 'file-checklist', 'index': dash.dependencies.ALL}, 'value'),
        prevent_initial_call=False
    )
    def update_data_access_table( n_intervals, selected_files):
        
        to_get_node_values = get_nodes_info(selected_files=selected_files)
        data = get_data_to_update_into_opcua_table(app=app, to_get_node_values=to_get_node_values)
        
        return opcua_components.data_access_view_table(data=data)

    @app.callback(
        dash.Output("data_access_view_table", "children"),
        dash.Input({'type': 'file-checklist', 'index': dash.dependencies.ALL}, 'value')
    )
    def display_selected_file(selected_files):

        subscription_handler.unsubscribe_all()
        to_get_node_values = get_nodes_info(selected_files=selected_files)

        data = list()
        subscriptions = dict()
        for client_name, namespaces in to_get_node_values.items():
            
            client = app.automation.get_opcua_client(client_name=client_name)
            if client:
                subscriptions[client_name] = client.create_subscription(1000, subscription_handler)
                infos = app.automation.get_node_attributes(client_name=client_name, namespaces=namespaces)
                if infos:
                    for info in infos:
                        _info = info[0]
                        namespace = _info["Namespace"]
                        data.append(
                            {
                                "server": client_name,
                                "namespace": namespace,
                                "data_type": _info["DataType"],
                                "display_name": _info["DisplayName"],
                                "value": _info["Value"],
                                "source_timestamp": _info["DataValue"].SourceTimestamp,
                                "status_code": _info["DataValue"].StatusCode.name
                            }
                        )
                        
                        node_id = client.get_node_id_by_namespace(namespace)
                        subscription = subscriptions[client_name]
                        subscription_handler.subscribe(subscription=subscription, client_name=client_name, node_id=node_id)

        return opcua_components.data_access_view_table(data=data)

    @app.callback(
        dash.Output("add_server_modal", "is_open"),
        dash.Input("add_server_button", "n_clicks"),
        [dash.State("add_server_modal", "is_open")],
    )
    def add_server_button(n, is_open):
        r"""
        Documentation here
        """
        if n:

            return not is_open
        
        return is_open
    
    @app.callback(
        dash.Output("remove_server_modal", "is_open"),
        dash.Input("remove_server_button", "n_clicks"),
        [dash.State("remove_server_modal", "is_open")],
    )
    def remove_server_button(n, is_open):
        r"""
        Documentation here
        """
        if n:
            clients = list()
            for client_name in app.automation.get_opcua_clients().keys():

                clients.append(
                    {"label": client_name, "value": client_name}
                )

            dash.set_props("opcua_client_names_options", {"options": clients})
            return not is_open
        
        return is_open
    
    @app.callback(
        dash.Output("add_server_modal", "is_open", allow_duplicate=True),
        dash.Output("server_tree", "children"),
        dash.Input("add_server_ok_button_modal", "n_clicks"),
        dash.State("opcua_client_name_input", "value"),
        dash.State("opcua_client_host_input", "value"),
        dash.State("opcua_client_port_input", "value")
    )
    def ok_add_server_button(n, client_name:str, host:str="127.0.0.1", port:int=4840):
        r"""
        Documentation here
        """
        resp, _ = app.automation.add_opcua_client(client_name=client_name, host=host, port=port)
        if resp:

            data = OPCUAComponents.get_opcua_tree(app)
            subscription_handler.unsubscribe_all()        

            return False, data
        
        message = f"Connection refused on opc.tcp://{host}:{port}"
        dash.set_props("modal-error-opcua-connection-body", {"children": message})
        dash.set_props("modal-error-opcua-connection", {'is_open': True})

        return False, {}
    
    @app.callback(
        dash.Output("add_server_modal", "is_open", allow_duplicate=True),
        dash.Input("add_server_cancel_button_modal", "n_clicks"),
    )
    def cancel_add_server_button(n):
        r"""
        Documentation here
        """
        return False
    
    @app.callback(
        dash.Output("remove_server_modal", "is_open", allow_duplicate=True),
        dash.Input("remove_server_ok_button_modal", "n_clicks"),
        dash.State("opcua_client_names_options", "value")
    )
    def ok_remove_server_button(n, client_name:str):
        r"""
        Documentation here
        """
        if client_name:

            app.automation.remove_opcua_client(client_name=client_name)

        data = OPCUAComponents.get_opcua_tree(app)
        dash.set_props("server_tree", {"children": data})
        return False
    
    @app.callback(
        dash.Output("remove_server_modal", "is_open", allow_duplicate=True),
        dash.Input("remove_server_cancel_button_modal", "n_clicks"),
    )
    def cancel_remove_server_button(n):
        r"""
        Documentation here
        """
        return False
    
    @app.callback(
        dash.Output("server_tree", "children", allow_duplicate=True),
        dash.Input('communications_page', 'pathname'),
        prevent_initial_call=True
        )
    def display_page(pathname):
        r"""
        Documentation here
        """
        if pathname=="/":

            data = OPCUAComponents.get_opcua_tree(app)
            subscription_handler.unsubscribe_all()
            
            return data
        
    @app.callback(
        dash.Output("modal-success-opcua-connection", "is_open"),
        dash.Input("close-success-opcua-connection", "n_clicks"),
        [dash.State("modal-success-opcua-connection", "is_open")],
    )
    def close_success_button(n, is_open):
        r"""
        Documentation here
        """
        if n:

            return not is_open
        
        return is_open
    
    @app.callback(
        dash.Output("modal-error-opcua-connection", "is_open"),
        dash.Input("close-error-opcua-connection", "n_clicks"),
        [dash.State("modal-error-opcua-connection", "is_open")],
    )
    def close_error_opcua_connection_button(n, is_open):
        r"""
        Documentation here
        """
        if n:

            return not is_open
        
        return is_open
    
    @app.callback(
        dash.Output("download_node_info_modal", "is_open"),
        dash.Input("download_node_info_button", "n_clicks"),
        [dash.State("download_node_info_modal", "is_open")],
    )
    def download_node_info_button(n, is_open):
        r"""
        Documentation here
        """
        if n:
            clients = list()
            for client_name in app.automation.get_opcua_clients().keys():

                clients.append(
                    {"label": client_name, "value": client_name}
                )

            dash.set_props("download_opcua_client_names_options", {"options": clients})
            return not is_open
        
        return is_open
    
    @app.callback(
        dash.Output("download_node_info_modal", "is_open", allow_duplicate=True),
        dash.Output("download", "data"),
        dash.Input("download_node_info_ok_button_modal", "n_clicks"),
        dash.State("download_opcua_client_names_options", "value")
    )
    def ok_download_node_info_button(n, client_name:str):
        r"""
        Documentation here
        """
        if client_name:

            opcua_tree = app.automation.get_opcua_tree(client_name=client_name)
            opcua_tree = opcua_tree[0]['Objects']
            flat = list()
            for tree in opcua_tree:
                
                flat.extend(OPCUAComponents.flatten_dict(tree))
            output = io.StringIO() 
            writer = csv.DictWriter(output, fieldnames=["name", "namespace", "NodeClass"]) 
            writer.writeheader() 
            writer.writerows(flat)
            
        return False, dict(content=output.getvalue(), filename=f"opcua_server_node_{client_name}.csv")
    
    @app.callback(
        dash.Output("download_node_info_modal", "is_open", allow_duplicate=True),
        dash.Input("download_node_info_cancel_button_modal", "n_clicks"),
    )
    def cancel_remove_server_button(n):
        r"""
        Documentation here
        """
        return False


from ....extensions.api import api


def init_app():

    from .alarms import ns as ns_alarms
    from .summary import ns as ns_summary

    api.add_namespace(ns_alarms, path="/alarms")
    api.add_namespace(ns_summary, path="/alarms/summary")


# -*- coding: utf-8 -*-
"""pyhades/managers/alarms.py
This module implements Alarm Manager.
"""
from datetime import datetime
import queue
from ..singleton import Singleton
from ..tags import CVTEngine, TagObserver
from ..alarms import AlarmState, Alarm
from ..dbmodels.alarms import AlarmSummary
from ..modules.users.users import User
from ..models import FloatType, StringType
from ..utils.decorators import set_event, logging_error_handler
from flask_socketio import SocketIO


class AlarmManager(Singleton):
    r"""
    This class implements all definitions for the Alarm Management System
    """

    def __init__(self):

        self._alarms:dict[Alarm] = dict()
        self._tag_queue = queue.Queue()
        self.tag_engine = CVTEngine()

    def get_queue(self)->queue.Queue:
        r"""
        Documentation here
        """
        return self._tag_queue

    @logging_error_handler
    def append_alarm(
            self,
            name:str,
            tag:str,
            type:str="BOOL",
            trigger_value:bool|float=True,
            description:str="",
            identifier:str=None,
            state:str="Normal",
            timestamp:str=None,
            ack_timestamp:str=None,
            user:User=None,
            reload:bool=False,
            sio:SocketIO|None=None
        )->tuple[Alarm, str]:
        r"""
        Append alarm to the Alarm Manager

        **Paramters**

        * **alarm**: (Alarm Object)

        **Returns**

        * **None**
        """
        # Check alarm name duplicated
        alarm = self.get_alarm_by_name(name)
        if alarm:

            return alarm, f"Alarm {name} is already defined"

        # Check if alarm is associated to same tag with same alarm type
        trigger_value_message = self.__check_trigger_values(name=name, tag=tag, type=type, trigger_value=trigger_value)
        if trigger_value_message:

            return None, trigger_value_message

        if timestamp:

            timestamp = datetime.strptime(timestamp, self.tag_engine.DATETIME_FORMAT)

        if ack_timestamp:

            ack_timestamp = datetime.strptime(ack_timestamp, self.tag_engine.DATETIME_FORMAT)

        alarm = Alarm(
            name=name,
            tag=self.tag_engine.get_tag_by_name(name=tag),
            description=description,
            alarm_type=StringType(type),
            alarm_setpoint=FloatType(trigger_value),
            identifier=identifier,
            state=state,
            timestamp=timestamp,
            ack_timestamp=ack_timestamp,
            user=user,
            reload=reload
        )
        alarm.set_socketio(sio=sio)
        self._alarms[alarm.identifier] = alarm

        return alarm, f"Alarm creation successful"

    @logging_error_handler
    def put(
            self,
            id:str,
            name:str=None,
            tag:str=None,
            description:str=None,
            alarm_type:str=None,
            trigger_value:float=None,
            user:User=None
            )->tuple[Alarm, str]:
        r"""
        Updates alarm attributes

        **Parameters**

        * **id** (int).
        * **name** (str)[Optional]:
        * **tag** (str)[Optional]:
        * **description** (str)[Optional]:
        * **type** (str)[Optional]:
        * **trigger** (float)[Optional]:

        **Returns**

        * **alarm** (dict) Alarm Object jsonable
        """
        alarm = self.get_alarm(id=id)
        if name:

            if self.get_alarm_by_name(name=name):

                return f"Alarm {name} is already defined"

        # Check if alarm is associated to same tag with same alarm type
        if not tag:
            tag = alarm.tag
        if not alarm_type:
            alarm_type = alarm.alarm_setpoint.type
        if not trigger_value:
            trigger_value = FloatType(alarm.alarm_setpoint.value)

        trigger_value_message = self.__check_trigger_values(
            name=alarm.name,
            tag=tag,
            type=alarm_type,
            trigger_value=trigger_value
            )
        if trigger_value_message:

            return None, trigger_value_message

        alarm, message = alarm.put(
            user=user,
            name=name,
            tag=tag,
            description=description,
            alarm_type=alarm_type,
            trigger_value=trigger_value
            )
        self._alarms[id] = alarm

    @logging_error_handler
    @set_event(message=f"Deleted", classification="Alarm", priority=3, criticity=5)
    def delete_alarm(self, id:str, user:User=None):
        r"""
        Removes alarm

        **Paramters**

        * **id** (int): Alarm ID
        """
        if id in self._alarms:

            alarm = self._alarms.pop(id)
            alarm.remove_from_service(user=user)

        return alarm, f"Alarm: {alarm.name} - Tag: {alarm.tag}"

    @logging_error_handler
    def get_alarm(self, id:str)->Alarm:
        r"""
        Gets alarm from the Alarm Manager by id

        **Paramters**

        * **id**: (int) Alarm ID

        **Returns**

        * **alarm** (Alarm Object)
        """

        if id in self._alarms:

            return self._alarms[id]

    @logging_error_handler
    def get_alarm_by_name(self, name:str)->Alarm:
        r"""
        Gets alarm from the Alarm Manager by name

        **Paramters**

        * **name**: (str) Alarm name

        **Returns**

        * **alarm** (Alarm Object)
        """
        for id, alarm in self._alarms.items():

            if name == alarm.name:

                return self._alarms[str(id)]

    @logging_error_handler
    def get_alarms_by_tag(self, tag:str)->dict:
        r"""
        Gets all alarms associated to some tag

        **Parameters**

        * **tag**: (str) tag name binded to alarm

        **Returns**

        * **alarm** (dict) of alarm objects
        """
        alarms = dict()
        for id, alarm in self._alarms.items():

            if tag == alarm.tag:

                alarms[id] = alarm

        return alarms

    @logging_error_handler
    def get_alarm_by_tag(self, tag:str)->list[Alarm]:
        r"""
        Gets alarm associated to some tag

        **Parameters**

        * **tag**: (str) tag name binded to alarm

        **Returns**

        * **alarm** (list) of alarm objects
        """
        alarms = list()
        for _, alarm in self._alarms.items():

            if tag == alarm.tag:

                alarms.append(alarm)

        return alarms

    @logging_error_handler
    def get_alarms(self)->dict:
        r"""
        Gets all alarms

        **Returns**

        * **alarms**: (dict) Alarm objects
        """
        return self._alarms

    @logging_error_handler
    def get_lasts_active_alarms(self, lasts:int=None)->list:
        r"""
        Documentation here
        """
        original_list = [alarm.serialize() for _, alarm in self.get_alarms().items()]
        filtered_list = [elem for elem in original_list if elem['state']['alarm_status'].lower()=="active"]
        sorted_list = sorted(filtered_list, key=lambda x: x['timestamp'] if x['timestamp'] else '')
        if lasts:

            if len(sorted_list)>lasts:

                sorted_list = sorted_list[0:lasts]

        return sorted_list

    @logging_error_handler
    def serialize(self)->list:
        r"""
        Documentation here
        """

        return [alarm.serialize() for _, alarm in self._alarms.items()]

    @logging_error_handler
    def get_tag_alarms(self)->list:
        r"""
        Gets all tag alarms defined

        **Returns**

        * **tags_alarms**: (list) alarm tags
        """
        result = [_alarm.tag_alarm for id, _alarm in self.get_alarms().items()]

        return result

    @logging_error_handler
    def tags(self)->list:
        r"""
        Gets all tags variables binded into alarms

        **Returns**

        * **tags**: (list)
        """
        result = set([_alarm.tag for id, _alarm in self.get_alarms().items()])

        return list(result)

    @logging_error_handler
    def __check_trigger_values(self, name:str, tag:str, type:str, trigger_value:float)->None|str:
        r"""
        Documentation here
        """
        alarms = self.get_alarm_by_tag(tag=tag)

        if alarms:

            for alarm in alarms:

                if alarm.name!=name:

                    if type==alarm.alarm_setpoint.type.value:

                        return f"Alarm Type {type} and alarm's tag {tag} duplicated"

                    if type=="LOW-LOW":

                        if trigger_value>=alarm.alarm_setpoint.value:

                            return f"Conflict definition with {alarm.name} in trigger value {trigger_value}>={alarm.alarm_setpoint.value}"

                    if type=="LOW":

                        if alarm.alarm_setpoint.type.value=="LOW-LOW":

                            if trigger_value<=alarm.alarm_setpoint.value:

                                return f"Conflict definition with {alarm.name} in trigger value {trigger_value}>={alarm.alarm_setpoint.value}"

                        else:

                            if trigger_value>=alarm.alarm_setpoint.value:

                                return f"Conflict definition with {alarm.name} in trigger value {trigger_value}>={alarm.alarm_setpoint.value}"

                    if type=="HIGH":

                        if alarm.alarm_setpoint.type.value=="HIGH-HIGH":

                            if trigger_value>=alarm.alarm_setpoint.value:

                                return f"Conflict definition with {alarm.name} in trigger value {trigger_value}<={alarm.alarm_setpoint.value}"

                        else:

                            if trigger_value<=alarm.alarm_setpoint.value:

                                return f"Conflict definition with {alarm.name} in trigger value {trigger_value}<={alarm.alarm_setpoint.value}"

                    if type=="HIGH-HIGH":

                        if trigger_value<=alarm.alarm_setpoint.value:

                            return f"Conflict definition with {alarm.name} in trigger value {trigger_value}<={alarm.alarm_setpoint.value}"

    @logging_error_handler
    def filter_by(self, **fields):
        r"""
        Documentation here
        """

        return AlarmSummary.filter_by(**fields), 200

    @logging_error_handler
    def get_lasts(self, lasts:int=10):
        r"""
        Documentation here
        """

        return AlarmSummary.read_lasts(lasts=lasts), 200

    @logging_error_handler
    def summary(self)->dict:
        r"""
        Summarizes all Alarm Manager

        **Returns**

        * **summary**: (dict)
        """
        result = dict()
        alarms = [_alarm.name for id, _alarm in self.get_alarms().items()]
        result["length"] = len(alarms)
        result["alarms"] = alarms
        result["alarm_tags"] = self.get_tag_alarms()
        result["tags"] = self.tags()

        return result

    @logging_error_handler
    def attach(self, alarm_name:str):

        def attach_observer(entity):

            _tag = entity.tag

            observer = TagObserver(self._tag_queue)
            self.tag_engine.attach(name=_tag, observer=observer)

        alarm = self.get_alarm_by_name(name=alarm_name)
        attach_observer(alarm)

    @logging_error_handler
    def execute(self, tag_name:str):
        r"""
        Execute update state value of alarm if the value store in cvt for tag
        reach alarm threshold values

        **Paramters**

        * **tag**: (str) Tag in CVT
        """
        value = self.tag_engine.get_value_by_name(tag_name=tag_name)['value']

        for _, _alarm in self._alarms.items():

            if _alarm.state == AlarmState.SHLVD:

                _now = datetime.now()

                if _alarm._shelved_until:

                    if _now >= _alarm._shelved_until:

                        _alarm.unshelve()
                        continue

                    continue

                continue

            if tag_name==_alarm.tag:

                _alarm.update(value)



import dash
import dash_bootstrap_components as dbc

class AlarmSummaryComponents:

    @classmethod
    def alarm_summary_table(cls)->dash.dash_table.DataTable:
        r"""
        Documentation here
        """
        return dbc.Container(
            dbc.Row(
                dbc.Col(
                    dash.dash_table.DataTable(
                        data=[],
                        columns=[
                            {'name': 'id', 'id': 'id', 'editable': False}, 
                            {'name': 'name', 'id': 'name', 'editable': False}, 
                            {'name': 'tag', 'id': 'tag', 'editable': False}, 
                            {'name': 'description', 'id': 'description', 'editable': False}, 
                            {'name': 'state', 'id': 'state', 'editable': False}, 
                            {'name': 'alarm_time', 'id': 'alarm_time', 'editable': False}, 
                            {'name': 'ack_time', 'id': 'ack_time', 'editable': False}
                        ],
                        id="alarms_summary_datatable",
                        filter_action="native",
                        sort_action="native",
                        sort_mode="multi",
                        selected_columns=[],
                        page_action="native",
                        page_current= 0,
                        page_size= 10,
                        persistence=True,
                        editable=True,
                        persisted_props=['data'],
                        export_format='xlsx',
                        export_headers='display',
                        style_table={'overflowX': 'auto'},
                    ),
                    width=12,
                )
            ),
            fluid=True,
            className="mx-0 px-0"
        )


import secrets, logging
from datetime import datetime
from ..utils import Observer
from ..utils.decorators import logging_error_handler
from ..buffer import Buffer
from ..variables import (
    Temperature,
    Length,
    Current,
    Time,
    Pressure,
    Mass,
    Force,
    Power,
    VolumetricFlow,
    MassFlow,
    Density,
    Percentage,
    Adimentional,
    Volume
)
from .filter import GaussianFilter

DATETIME_FORMAT = "%m/%d/%Y, %H:%M:%S.%f"

class Tag:

    def __init__(
            self,
            name:str,
            unit:str,
            variable:str,
            data_type:str,
            display_name:str=None,
            display_unit:str=None,
            description:str="",
            opcua_address:str=None,
            node_namespace:str=None,
            scan_time:int=None,
            dead_band:float=None,
            timestamp:datetime=None,
            process_filter:bool=False,
            gaussian_filter:bool=False,
            gaussian_filter_threshold:float=1.0,
            gaussian_filter_r_value:float=0.0,
            outlier_detection:bool=False,
            out_of_range_detection:bool=False,
            frozen_data_detection:bool=False,
            manufacturer:str="",
            segment:str="",
            id:str=None
    ):
        self.id = secrets.token_hex(4)
        if id:
            self.id = id
        self.name = name
        self.data_type = data_type
        self.description = description
        self.variable = variable
        self.display_name = name
        if display_name:
            self.display_name = display_name
        self.display_unit = unit
        if display_unit:
            self.display_unit = display_unit
        self.unit=unit
        if variable.lower()=="temperature":
            self.value = Temperature(value=0.0, unit=self.unit)
        elif variable.lower()=="length":
            self.value = Length(value=0.0, unit=self.unit)
        elif variable.lower()=="time":
            self.value = Time(value=0.0, unit=self.unit)
        elif variable.lower()=="pressure":
            self.value = Pressure(value=0.0, unit=self.unit)
        elif variable.lower()=="mass":
            self.value = Mass(value=0.0, unit=self.unit)
        elif variable.lower()=="force":
            self.value = Force(value=0.0, unit=self.unit)
        elif variable.lower()=="power":
            self.value = Power(value=0.0, unit=self.unit)
        elif variable.lower()=="current":
            self.value = Current(value=0.0, unit=self.unit)
        elif variable.lower()=="volumetricflow":
            self.value = VolumetricFlow(value=0.0, unit=self.unit)
        elif variable.lower()=="massflow":
            self.value = MassFlow(value=0.0, unit=self.unit)
        elif variable.lower()=="density":
            self.value = Density(value=0.0, unit=self.unit)
        elif variable.lower()=="percentage":
            self.value = Percentage(value=0.0, unit=self.unit)
        elif variable.lower()=="adimentional":
            self.value = Adimentional(value=0.0, unit=self.unit)
        elif variable.lower()=="volume":
            self.value = Volume(value=0.0, unit=self.unit)

        self.values = Buffer()
        self.timestamps = Buffer()
        self.opcua_address = opcua_address
        self.node_namespace = node_namespace
        self.scan_time = scan_time
        self.dead_band = dead_band
        self.timestamp = timestamp
        self.process_filter = process_filter
        self.gaussian_filter = gaussian_filter
        self.gaussian_filter_threshold = gaussian_filter_threshold
        self.gaussian_filter_r_value = gaussian_filter_r_value
        self.outlier_detection = outlier_detection
        self.out_of_range_detection = out_of_range_detection
        self.frozen_data_detection = frozen_data_detection
        self.manufacturer = manufacturer
        self.segment = segment
        self.filter = GaussianFilter()
        self._observers = set()

    def set_name(self, name:str):
        r"""
        Documentation here
        """
        self.name = name

    @logging_error_handler
    def set_value(self, value:float|str|int|bool, timestamp:datetime=None):
        r"""
        Documentation here
        """
        if self.dead_band and isinstance(value, (int, float)):
            try:
                current_value = self.value.value
                if abs(value - current_value) < self.dead_band:
                    
                    return
            except Exception as e:
                logging.error(f"Error in deadband logic: {e}")

        if not timestamp:
            timestamp = datetime.now()
        self.value.set_value(value=value, unit=self.display_unit)
        self.timestamp = timestamp
        self.values(self.get_value())
        self.timestamps(timestamp.strftime(DATETIME_FORMAT))
        self.notify()

    def set_display_name(self, name:str):
        r"""
        Documentation here
        """

        self.display_name = name

    def set_data_type(self, data_type:str):
        r"""
        Documentation here
        """
        self.data_type = data_type

    def set_variable(self, variable:str):
        r"""
        Documentation here
        """

        self.variable = variable
        if variable.lower()=="temperature":
            self.value = Temperature(value=0.0, unit=self.unit)
        elif variable.lower()=="length":
            self.value = Length(value=0.0, unit=self.unit)
        elif variable.lower()=="time":
            self.value = Time(value=0.0, unit=self.unit)
        elif variable.lower()=="pressure":
            self.value = Pressure(value=0.0, unit=self.unit)
        elif variable.lower()=="mass":
            self.value = Mass(value=0.0, unit=self.unit)
        elif variable.lower()=="force":
            self.value = Force(value=0.0, unit=self.unit)
        elif variable.lower()=="power":
            self.value = Power(value=0.0, unit=self.unit)
        elif variable.lower()=="current":
            self.value = Current(value=0.0, unit=self.unit)
        elif variable.lower()=="volumetricflow":
            self.value = VolumetricFlow(value=0.0, unit=self.unit)
        elif variable.lower()=="massflow":
            self.value = MassFlow(value=0.0, unit=self.unit)
        elif variable.lower()=="density":
            self.value = Density(value=0.0, unit=self.unit)
        elif variable.lower()=="percentage":
            self.value = Percentage(value=0.0, unit=self.unit)
        elif variable.lower()=="adimentional":
            self.value = Adimentional(value=0.0, unit=self.unit)
        elif variable.lower()=="volume":
            self.value = Volume(value=0.0, unit=self.unit)

    def set_opcua_address(self, opcua_address:str):
        r"""
        Documentation here
        """
        self.opcua_address = opcua_address

    def set_unit(self, unit:str):
        r"""
        Documentation here
        """
        self.unit = unit

    def set_display_unit(self, unit:str): 
        r"""
        Documentation here
        """
        self.display_unit = unit

    def set_node_namespace(self, node_namespace:str):
        r"""
        Documentation here
        """
        self.node_namespace = node_namespace

    def get_value(self):
        r"""
        Documentation here
        """            
        return round(self.value.convert(to_unit=self.display_unit), 3)
    
    def set_description(self, description:str):
        r"""
        Documentation here
        """
        self.description = description
    
    def set_scan_time(self, scan_time:int):
        r"""
        Documentation here
        """
        self.scan_time = scan_time

    def set_dead_band(self, dead_band:float):
        r"""
        Documentation here
        """
        self.dead_band = dead_band

    def get_timestamp(self):
        r"""
        Documentation here
        """
        return self.timestamp

    def get_scan_time(self):
        r"""
        Documentation here
        """
        return self.scan_time
    
    def get_dead_band(self):
        r"""
        Documentation here
        """
        return self.dead_band

    def get_data_type(self):
        r"""
        Documentation here
        """
        return self.data_type

    def get_unit(self):
        r"""
        Documentation here
        """
        return self.unit
    
    def get_display_unit(self):
        r"""
        Documentation here
        """
        return self.display_unit

    def get_description(self):
        r"""
        Documentation here
        """
        return self.description
    
    def get_display_name(self)->str:
        r"""
        Documentation here
        """

        return self.display_name
    
    def get_variable(self)->str:
        r"""
        Documentation here
        """

        return self.variable
    
    def get_id(self)->str:
        r"""
        Documentation here
        """

        return self.id
    
    def get_name(self)->str:
        r"""
        Documentation here
        """

        return self.name

    def get_opcua_address(self):
        r"""
        Documentation here
        """
        return self.opcua_address

    def get_node_namespace(self):
        r"""
        Documentation here
        """
        return self.node_namespace
    
    def attach(self, observer:Observer):
        r"""
        Documentation here
        """
        observer._subject = self
        self._observers.add(observer)

    def detach(self, observer:Observer):
        r"""
        Documentation here
        """
        observer._subject = None
        self._observers.discard(observer)

    def notify(self):
        r"""
        Documentation here
        """
        for observer in self._observers:
            
            observer.update()

    def serialize(self):

        timestamp = self.get_timestamp()
        if timestamp:

            timestamp = timestamp.strftime(DATETIME_FORMAT)

        return {
            "id": self.get_id(),
            "value": self.get_value(),
            "timestamp": timestamp,
            "values": list(self.values),
            "timestamps": list(self.timestamps),
            "name": self.name,
            "unit": self.get_unit(),
            "display_unit": self.get_display_unit(),
            "data_type": self.get_data_type(),
            "variable": self.get_variable(),
            "description": self.get_description(),
            "display_name": self.get_display_name(),
            "opcua_address": self.get_opcua_address(),
            "node_namespace": self.get_node_namespace(),
            "scan_time": self.get_scan_time(),
            "dead_band": self.get_dead_band(),
            "segment": self.segment,
            "manufacturer": self.manufacturer,
            "process_filter": self.process_filter,
            "gaussian_filter": self.gaussian_filter,
            "out_of_range_detection": self.out_of_range_detection,
            "frozen_data_detection": self.frozen_data_detection,
            "outlier_detection": self.outlier_detection
        }


class TagObserver(Observer):
    """
    Implement the Observer updating interface to keep its state
    consistent with the subject's.
    Store state that should stay consistent with the subject's.
    """
    def __init__(self, tag_queue):

        super(TagObserver, self).__init__()
        self._tag_queue = tag_queue

    def update(self):

        """
        This methods inserts the changing Tag into a 
        Producer-Consumer Queue Design Pattern
        """
        result = dict()
        result["tag"] = self._subject.name
        result["value"] = self._subject.value.convert(self._subject.get_display_unit())
        result["timestamp"] = self._subject.timestamp
        self._tag_queue.put(result, block=False)


class MachineObserver(Observer):
    """
    Implement the Observer updating interface to keep its state
    consistent with the subject's.
    Store state that should stay consistent with the subject's.
    """
    def __init__(self, machine):

        super(MachineObserver, self).__init__()
        self.machine = machine

    def update(self):

        """
        This methods inserts the changing Tag into a 
        Producer-Consumer Queue Design Pattern
        """
        self.machine.notify(tag=self._subject.name, value=self._subject.value, timestamp=self._subject.timestamp)



# -*- coding: utf-8 -*-
"""pyhades/logger/logdict.py

This module implements a dictionary based Class to hold
the tags to be logged.
"""


class LogTable(dict):

    def __init__(self):

        pass

    def validate(self, period, tag):
        
        if not type(period) in [int, float]:
            return False
        
        if type(tag) != str:
            return False

        return True

    def get_groups(self):

        return list(self.keys())

    def get_tags(self, group):

        return self[group]

    def get_all_tags(self):

        result = list()

        for group in self.get_groups():

            result += self.get_tags(group)

        return result

    def get_period(self, tag):

        for key, value in self.items():

            if tag in value:
                return key

    def serialize(self):

        return self
    



from flask import Blueprint, request
from flask_restx import abort
from flask_restx import Api as API
from ..singleton import Singleton
from functools import wraps
import logging, jwt
from ..utils.decorators import decorator
from ..dbmodels.users import Users
from ..modules.users.users import Users as CVTUsers


authorizations = {
    'apikey' : {
        'type' : 'apiKey',
        'in' : 'header',
        'name' : 'X-API-KEY'
    }
}


blueprint = Blueprint('api', __name__, url_prefix='/api')

api = API(blueprint, version='1.0', 
        title='PyAutomation API',
        description="""
        This API groups all namespaces defined in every module's resources for PyAutomation App.
        """, 
        doc='/docs',
        authorizations=authorizations
    )

users = CVTUsers()

class Api(Singleton):

    def __init__(self):

        self.app = None

    def init_app(self, app):
        r"""
        Documentation here
        """
        self.app = self.create_api(app)

        return app

    def create_api(self, app):
        r"""
        Documentation here
        """
        app.register_blueprint(blueprint)

        return api
    
    @staticmethod
    def verify_tpt(tpt:str):
        r"""
        Verify Third Party Token
        """
        from .. import server
        try:

            jwt.decode(tpt, server.config["TPT_TOKEN"], algorithms=["HS256"])

            return True

        except:

            return 
        
    @classmethod
    def validate_reqparser(cls, reqparser):
        def _validate_reqparser(f):
                
            @wraps(f)
            def decorated(*args, **kwargs):
                
                reqparser.parse_args() 
                result = f(*args, **kwargs)
                return result

            return decorated
        
        return _validate_reqparser
    
    @classmethod
    def token_required(cls, auth:bool=False):
        
        def _token_required(f):
            
            @wraps(f)
            def decorated(*args, **kwargs):
                try:

                    if auth:

                        token = None

                        if 'X-API-KEY' in request.headers:
                            
                            token = request.headers['X-API-KEY']

                        elif 'Authorization' in request.headers:
                            
                            token = request.headers['Authorization'].split('Token ')[-1]

                        if not token:
                            
                            return {'message' : 'Key is missing.'}, 401
                        
                        user = Users.get_or_none(token=token)

                        if user:

                            return f(*args, **kwargs)

                        if Api.verify_tpt(tpt=token):
                    
                            return f(*args, **kwargs)

                        return {'message' : 'Invalid token'}, 401                  
                
                except Exception as err:
                    logger = logging.getLogger("pyautomation")
                    logger.error(str(err))

            return decorated

        return _token_required
    
    @classmethod
    def get_current_user(cls):

        token = None

        if 'X-API-KEY' in request.headers:
                            
            token = request.headers['X-API-KEY']

        elif 'Authorization' in request.headers:
            
            token = request.headers['Authorization'].split('Token ')[-1]

        if token:

            return users.get_active_user(token=token)

        return None


import pytz
from datetime import datetime, timedelta
from flask_restx import Namespace, Resource, fields
from .... import PyAutomation
from ....extensions.api import api
from ....extensions import _api as Api
from .... import _TIMEZONE, TIMEZONE

ns = Namespace('Operation Logs', description='Operation Logs')
app = PyAutomation()


logs_filter_model = api.model("logs_filter_model",{
    'usernames': fields.List(fields.String(), required=False),
    'alarm_names': fields.List(fields.String(), required=False),
    'event_ids': fields.List(fields.Integer(), required=False),
    'classification': fields.String(required=False),
    'message': fields.String(required=False),
    'description': fields.String(required=False),
    'greater_than_timestamp': fields.DateTime(required=False, default=datetime.now(pytz.utc).astimezone(TIMEZONE) - timedelta(minutes=30), description=f'Greater than timestamp - DateTime Format: {app.cvt.DATETIME_FORMAT}'),
    'less_than_timestamp': fields.DateTime(required=False, default=datetime.now(pytz.utc).astimezone(TIMEZONE), description=f'Less than timestamp - DateTime Format: {app.cvt.DATETIME_FORMAT}',),
    'timezone': fields.String(required=False, default=_TIMEZONE)
})

logs_model = api.model("logs_model",{
    'message': fields.String(required=True, description="Log message"),
    'alarm_summary_id': fields.Integer(required=False, description="Alarm summary id comment"),
    'event_id': fields.Integer(required=False, description="Event id comment"),
    'description': fields.String(required=False, description="Log description")
})

@ns.route('/add')
class AddLogsByResource(Resource):

    @api.doc(security='apikey')
    @Api.token_required(auth=True)
    @ns.expect(logs_model)
    def post(self):
        r"""
        Create Log
        """
        user = Api.get_current_user()
        api.payload.update({
            "user": user
        })
        if "event_id" in api.payload:
            api.payload.update({
                "classification": "Event"
            })
        elif "alarm_summary_id" in api.payload:
            api.payload.update({
                "classification": "Alarm"
            })
        else:
            api.payload.update({
                "classification": "General"
            })

        log, message = app.create_log(**api.payload)
        if log:
            
            return log.serialize(), 200
        
        return message, 400

    
@ns.route('/filter_by')
class LogsFilterByResource(Resource):

    @api.doc(security='apikey')
    @Api.token_required(auth=True)
    @ns.expect(logs_filter_model)
    def post(self):
        r"""
        Logs Filter By
        """
        timezone = _TIMEZONE
        if "timezone" in api.payload:

            timezone = api.payload["timezone"]

        if timezone not in pytz.all_timezones:

            return f"Invalid Timezone", 400
        
        separator = '.'
        if 'greater_than_timestamp' in api.payload:
            
            greater_than_timestamp = api.payload['greater_than_timestamp']
            api.payload['greater_than_timestamp'] = greater_than_timestamp.replace("T", " ").split(separator, 1)[0] + '.00'
        
        if "less_than_timestamp" in api.payload:

            less_than_timestamp = api.payload['less_than_timestamp']
            api.payload['less_than_timestamp'] = less_than_timestamp.replace("T", " ").split(separator, 1)[0] + '.00'
        return app.filter_logs_by(**api.payload), 200
    

@ns.route('/lasts/<lasts>')
class LastsEventsResource(Resource):

    @api.doc(security='apikey')
    @Api.token_required(auth=True)
    def get(self, lasts:int=10):
        r"""
        Get lasts events
        """
        
        return app.get_lasts_logs(lasts=int(lasts)), 200



# -*- coding: utf-8 -*-
"""pyhades/logger/machines.py
"""
from ..dbmodels import Machines, TagsMachines, Tags
from .core import BaseEngine, BaseLogger
from ..utils.decorators import db_rollback
from ..models import IntegerType, StringType, FloatType
from ..tags.tag import Tag


class MachinesLogger(BaseLogger):

    def __init__(self):

        super(MachinesLogger, self).__init__()

    @db_rollback
    def create(
            self,
            identifier:str,
            name:str,
            interval:int,
            description:str,
            classification:str,
            buffer_size:int,
            buffer_roll_type:str,
            criticity:int,
            priority:int,
            on_delay:int=None,
            threshold:float=None
            ):
        r"""
        Documentation here
        """
        if not self.check_connectivity():

            return None
        
        if hasattr(threshold, "value"):
            
            threshold = threshold.value
       
        Machines.create(
            identifier=identifier,
            name=name,
            interval=interval,
            description=description,
            classification=classification,
            buffer_size=buffer_size,
            buffer_roll_type=buffer_roll_type,
            criticity=criticity,
            priority=priority,
            on_delay=on_delay,
            threshold=threshold
        )

    @db_rollback
    def put(
        self,
        name:StringType,
        machine_interval:IntegerType=None,
        description:StringType=None,
        classification:StringType=None,
        buffer_size:IntegerType=None,
        buffer_roll_type:StringType=None,
        criticity:IntegerType=None,
        priority:IntegerType=None,
        on_delay:IntegerType=None,
        threshold:FloatType=None
        ):

        if not self.check_connectivity():
            
            return None

        fields = dict()
        machine = Machines.read_by_name(name=name.value)
        if machine_interval:
            
            fields["interval"] = machine_interval.value
        if description:
            fields["description"] = description.value
        if classification:
            fields["classification"] = classification.value
        if buffer_size:
            fields["buffer_size"] = buffer_size.value
        if buffer_roll_type:
            fields["buffer_roll_type"] = buffer_roll_type.value
        if criticity:
            fields["criticity"] = criticity.value
        if priority:
            fields["priority"] = priority.value
        if on_delay:
            fields["on_delay"] = on_delay.value
        if threshold:
            if hasattr(threshold.value, "value"):
                threshold.value = threshold.value.value
            fields["threshold"] = threshold.value
            
        query = Machines.put(
            id=machine.id,
            **fields
        )

        return query
    
    @db_rollback
    def read_all(self):
        if not self.check_connectivity():

            return list()
        
        return Machines.read_all()
    
    @db_rollback
    def read_config(self):
        if not self.check_connectivity():
            
            return None

        return Machines.read_config()
    
    @db_rollback
    def bind_tag(self, tag:Tag, machine, default_tag_name:str=None):
        if not self.check_connectivity():

            return None
            
        TagsMachines.create(tag_name=tag.name, machine_name=machine.name.value, default_tag_name=default_tag_name)

    @db_rollback
    def unbind_tag(self, tag:Tag, machine):
        if not self.check_connectivity():

            return None

        tag_from_db = Tags.get_or_none(name=tag.name)
        machine_from_db= Machines.get_or_none(name=machine.name.value)
        tags_machine = TagsMachines.get((TagsMachines.tag == tag_from_db) & (TagsMachines.machine == machine_from_db))
        tags_machine.delete_instance()    

class MachinesLoggerEngine(BaseEngine):
    r"""
    Alarms logger Engine class for Tag thread-safe database logging.

    """

    def __init__(self):

        super(MachinesLoggerEngine, self).__init__()
        self.logger = MachinesLogger()

    def create(
        self,
        identifier:str,
        name:str,
        interval:int,
        description:str,
        classification:str,
        buffer_size:int,
        buffer_roll_type:str,
        criticity:int,
        priority:int,
        on_delay:int=None,
        threshold:float=None
        ):

        _query = dict()
        _query["action"] = "create"
        _query["parameters"] = dict()
        _query["parameters"]["identifier"] = identifier
        _query["parameters"]["name"] = name
        _query["parameters"]["interval"] = interval
        _query["parameters"]["classification"] = classification
        _query["parameters"]["description"] = description
        _query["parameters"]["buffer_size"] = buffer_size
        _query["parameters"]["buffer_roll_type"] = buffer_roll_type
        _query["parameters"]["criticity"] = criticity
        _query["parameters"]["priority"] = priority
        _query["parameters"]["on_delay"] = on_delay
        _query["parameters"]["threshold"] = threshold
        
        return self.query(_query)
    
    def put(
        self,
        name:StringType,
        machine_interval:IntegerType=None,
        description:StringType=None,
        classification:StringType=None,
        buffer_size:IntegerType=None,
        buffer_roll_type:StringType=None,
        criticity:IntegerType=None,
        priority:IntegerType=None,
        on_delay:IntegerType=None,
        threshold:FloatType=None
        ):

        _query = dict()
        _query["action"] = "put"
        _query["parameters"] = dict()
        _query["parameters"]["name"] = name
        _query["parameters"]["machine_interval"] = machine_interval
        _query["parameters"]["classification"] = classification
        _query["parameters"]["description"] = description
        _query["parameters"]["buffer_size"] = buffer_size
        _query["parameters"]["buffer_roll_type"] = buffer_roll_type
        _query["parameters"]["criticity"] = criticity
        _query["parameters"]["priority"] = priority
        _query["parameters"]["on_delay"] = on_delay
        _query["parameters"]["threshold"] = threshold

        return self.query(_query)

    def read_all(self):

        _query = dict()
        _query["action"] = "read_all"
        _query["parameters"] = dict()
        return self.query(_query)
    
    def read_config(self):

        _query = dict()
        _query["action"] = "read_config"
        _query["parameters"] = dict()
        return self.query(_query)
    
    def bind_tag(self, tag:Tag, machine, default_tag_name:str=None):

        _query = dict()
        _query["action"] = "bind_tag"
        _query["parameters"] = dict()
        _query["parameters"]["tag"] = tag
        _query["parameters"]["machine"] = machine
        _query["parameters"]["default_tag_name"] = default_tag_name
        return self.query(_query)
    
    def unbind_tag(self, tag:Tag, machine):

        _query = dict()
        _query["action"] = "unbind_tag"
        _query["parameters"] = dict()
        _query["parameters"]["tag"] = tag
        _query["parameters"]["machine"] = machine
        return self.query(_query)
    




import secrets
from datetime import datetime, timedelta, timezone
from .states import AlarmState, AlarmAttrs
from .trigger import Trigger, TriggerType
from ..tags.tag import Tag, MachineObserver
from ..tags.cvt import CVTEngine
from ..modules.users.users import User
from ..utils.decorators import validate_types, logging_error_handler, set_event, put_alarm_state
from ..models import FloatType, IntegerType, StringType
from ..variables import *
from statemachine import State, StateMachine
from flask_socketio import SocketIO


class Alarm(StateMachine):

    # MAIN STATES
    normal = State("normal", initial=True)
    unack_alarm = State("unack_alarm")
    ack_alarm = State("ack_alarm")
    rtn_unack = State("rtn_unack")

    # OUT OF SERVICE STATES
    shelved = State("shelved")
    suppressed_by_design = State("suppressed_by_design")
    out_of_service = State("out_of_service")

    # MAIN TRANSITIONS
    normal_to_unack_alarm = normal.to(unack_alarm)
    unack_alarm_to_ack_alarm = unack_alarm.to(ack_alarm)
    ack_alarm_to_normal = ack_alarm.to(normal)
    unack_alarm_to_rtn_unack = unack_alarm.to(rtn_unack)
    rtn_unack_to_normal = rtn_unack.to(normal)
    rtn_unack_to_unack_alarm = rtn_unack.to(unack_alarm)

    # SHELVED TRANSITIONS
    normal_to_shelved = normal.to(shelved)
    unack_alarm_to_shelved = unack_alarm.to(shelved)
    ack_alarm_to_shelved = ack_alarm.to(shelved)
    rtn_unack_to_shelved = rtn_unack.to(shelved)
    shelved_to_normal = shelved.to(normal)
    shelved_to_unack_alarm = shelved.to(unack_alarm)

    # SUPPRESSED BY DESIGN TRANSITIONS
    normal_to_suppressed_by_design = normal.to(suppressed_by_design)
    unack_alarm_to_suppressed_by_design = unack_alarm.to(suppressed_by_design)
    ack_alarm_to_suppressed_by_design = ack_alarm.to(suppressed_by_design)
    rtn_unack_to_suppressed_by_design = rtn_unack.to(suppressed_by_design)
    suppressed_by_design_to_normal = suppressed_by_design.to(normal)
    suppressed_by_design_to_unack_alarm = suppressed_by_design.to(unack_alarm)


    # OUT OF SERVICE TRANSITIONS
    normal_to_out_of_service = normal.to(out_of_service)
    unack_alarm_to_out_of_service = unack_alarm.to(out_of_service)
    ack_alarm_to_out_of_service = ack_alarm.to(out_of_service)
    rtn_unack_to_out_of_service = rtn_unack.to(out_of_service)
    out_of_service_to_normal = out_of_service.to(normal)
    out_of_service_to_unack_alarm = out_of_service.to(unack_alarm)

    def __init__(
            self,
            name:str, 
            tag:Tag,
            alarm_type:StringType,
            alarm_setpoint:IntegerType|FloatType,
            description:str="",
            state:str=None,
            timestamp:datetime=None,
            ack_timestamp:datetime=None,
            alarm_deadband:IntegerType|FloatType=FloatType(0.0),
            alarm_on_delay:IntegerType|FloatType=FloatType(0.0),
            alarm_off_delay:IntegerType|FloatType=FloatType(0.0),
            identifier:str=None,
            user:User=None,
            reload:bool=False
        ):
        from ..logger.alarms import AlarmsLoggerEngine
        self.alarm_engine = AlarmsLoggerEngine()
        self.tag_engine = CVTEngine()
        self.name = name
        self.tag = tag
        self.segment = tag.segment
        self.manufacturer = tag.manufacturer
        self.attach(machine=self, tag=tag)
        self.description = description        
        self.alarm_setpoint = Trigger()
        self.alarm_setpoint.value = alarm_setpoint.value
        self.alarm_setpoint.type = TriggerType(value=alarm_type.value.upper())
        alarm_deadband.unit = tag.get_display_unit()
        self.alarm_deadband = alarm_deadband
        self.alarm_on_delay = alarm_on_delay
        self.alarm_off_delay = alarm_off_delay
        self.timestamp = timestamp 
        self.ack_timestamp = ack_timestamp
        self.state = AlarmState.NORM
        if state:
            
            for _, attr in AlarmState.__dict__.items():
                
                if isinstance(attr, AlarmAttrs):
                    
                    if state==attr.state:
                        
                        self.state = attr
                        break
        if identifier:
            self.identifier = identifier
        else:
            self.identifier = secrets.token_hex(4)

        self._shelved_time:datetime = None
        self._shelved_until:datetime = None
        self._shelved_options_time = {
            'days': 0,
            'seconds': 0,
            'microseconds': 0,
            'milliseconds': 0,
            'minutes': 0,
            'hours': 0,
            'weeks': 0
        }
        transitions = []
        for state in self.states:
            transitions.extend(state.transitions)
        self.transitions = transitions
        self.sio:SocketIO|None = None
        super(Alarm, self).__init__()

    @logging_error_handler
    @put_alarm_state
    def on_enter_normal(self):
        self.state = AlarmState.NORM
        self.timestamp = None 
        self.ack_timestamp = None

    @logging_error_handler
    @put_alarm_state
    def on_enter_unack_alarm(self):

        self.state = AlarmState.UNACK
        self.timestamp = self.__timestamp
        self.alarm_engine.create_record_on_alarm_summary(
                name=self.name, 
                state=self.state.state, 
                timestamp=self.timestamp,
                ack_timestamp=self.ack_timestamp
            )

    @logging_error_handler
    @put_alarm_state
    def on_enter_ack_alarm(self):

        self.state = AlarmState.ACKED
        self.ack_timestamp = self.__timestamp
        self.alarm_engine.put_record_on_alarm_summary(
            name=self.name, 
            state=self.state.state, 
            ack_timestamp=self.ack_timestamp
        )
    
    @logging_error_handler
    @put_alarm_state
    def on_enter_rtn_unack(self):
        
        self.timestamp = None
        self.state = AlarmState.RTNUN
        self.alarm_engine.put_record_on_alarm_summary(
            name=self.name, 
            state=self.state.state
        )
    
    @logging_error_handler
    @put_alarm_state
    def on_enter_shelved(self):
        
        self.state = AlarmState.SHLVD

    @logging_error_handler
    @put_alarm_state
    def on_enter_suppressed_by_design(self):
        
        self.state = AlarmState.DSUPR

    @logging_error_handler
    @put_alarm_state
    def on_enter_out_of_service(self):
        
        self.state = AlarmState.OOSRV

    def set_socketio(self, sio:SocketIO):
        r"""
        Documentation here
        """
        self.sio:SocketIO = sio

    @logging_error_handler
    @validate_types(
            tag=str, 
            value=Temperature|Length|Current|Time|Pressure|Mass|Force|Power|VolumetricFlow|MassFlow|Density|Percentage|Adimentional, 
            timestamp=datetime, 
            output=None)
    def notify(
        self, 
        tag:str, 
        value:Temperature|Length|Current|Time|Pressure|Mass|Force|Power|VolumetricFlow|MassFlow|Density|Percentage|Adimentional, 
        timestamp:datetime):
        r"""
        This method provide an interface to CVT to notify if tag value has change
        
        # Parameters

        - *tag:* [Tag] tag Object
        - *value:* [int|float|bool] tag value
        """ 
        self.__timestamp = timestamp
        if self.state not in (AlarmState.DSUPR, AlarmState.SHLVD, AlarmState.OOSRV):     
            if self.alarm_setpoint.type in (TriggerType.HH, TriggerType.H):

                if value.value > self.alarm_setpoint.value:

                    self.abnormal_condition()
                
                else: 

                    self.normal_condition()

            elif self.alarm_setpoint.type in (TriggerType.L, TriggerType.LL):

                if value.value < self.alarm_setpoint.value:

                    self.abnormal_condition()

                else:

                    self.normal_condition()

            else: # Boolean Alarm
                
                if value.value == bool(self.alarm_setpoint.value):

                    self.abnormal_condition()

                else:

                    self.normal_condition()

        if self.state==AlarmState.SHLVD:

            if datetime.now(timezone.utc) >= self._shelved_until:

                self.unshelve()
        
    @logging_error_handler
    def abnormal_condition(self):
        r"""
        Documentation here
        """
        current_state = self.current_state.name.lower()
        transition_name = f'{current_state}_to_unack_alarm'
        self.__transition(transition_name=transition_name)

    @logging_error_handler
    def normal_condition(self):
        r"""
        Documentation here
        """
        current_state = self.current_state.name.lower()

        if current_state=="unack_alarm":

            transition_name = f'{current_state}_to_rtn_unack'
            self.__transition(transition_name=transition_name)

        elif current_state=="ack_alarm":

            transition_name = f'{current_state}_to_normal'
            self.__transition(transition_name=transition_name)

    @logging_error_handler
    @set_event(message=f"Acknowledged", classification="Alarm", priority=2, criticity=3)
    def acknowledge(self, user:User=None):
        r"""
        Documentation here
        """
        current_state = self.current_state.name.lower()

        if current_state=="unack_alarm":

            transition_name = f'{current_state}_to_ack_alarm'

        elif current_state=="rtn_unack":

            transition_name = f'{current_state}_to_normal'


        tag = self.tag_engine.get_tag_by_name(name=self.tag.name)
        self.alarm_engine.put_record_on_alarm_summary(
            name=self.name, 
            ack_timestamp=tag.get_timestamp()
        )
        self.__transition(transition_name=transition_name)
        return self, f"{self.tag.get_name()}"

    @logging_error_handler
    @set_event(message=f"Shelved", classification="Alarm", priority=2, criticity=3)
    def shelve(self, user:User=None, **options):
        r"""
        Documentation here
                **Parameters**

        * **days:** (int)
        * **seconds:** (int)
        * **minutes:** (int)
        * **hours:** (int)
        * **weeks:** (int)
        """
        options_time = {key: options[key] if key in options else self._shelved_options_time[key] for key in self._shelved_options_time}
        
        if options_time!=self._shelved_options_time:
            
            self._shelved_time = datetime.now(timezone.utc)
            self._shelved_until = self._shelved_time + timedelta(**options_time)

        current_state = self.current_state.name.lower()
        transition_name = f'{current_state}_to_shelved'
        self.__transition(transition_name=transition_name)
        return self, f"{self.tag.get_name()}"

    @logging_error_handler
    @set_event(message=f"Unshelved", classification="Alarm", priority=2, criticity=3)
    def unshelve(self, user:User=None):
        r"""
        Documentation here
        """
        self.__return_to_service()
        return self, f"{self.tag.get_name()}"

    @logging_error_handler
    @set_event(message=f"Designed suppression", classification="Alarm", priority=2, criticity=3)
    def designed_suppression(self, user:User=None):
        r"""
        Documentation here
        """
        current_state = self.current_state.name.lower()
        transition_name = f'{current_state}_to_suppressed_by_design'
        self.__transition(transition_name=transition_name)
        return self, f"{self.tag.get_name()}"

    @logging_error_handler
    @set_event(message=f"Designed unsuppression", classification="Alarm", priority=2, criticity=3)
    def designed_unsuppression(self, user:User=None):
        r"""
        Documentation here
        """
        self.__return_to_service()
        return self, f"{self.tag.get_name()}"

    @logging_error_handler
    @set_event(message=f"Removed from service", classification="Alarm", priority=2, criticity=3)
    def remove_from_service(self, user:User=None):
        r"""
        Documentation here
        """
        current_state = self.current_state.name.lower()
        transition_name = f'{current_state}_to_out_of_service'
        self.__transition(transition_name=transition_name)
        return self, f"{self.tag.get_name()}"

    @logging_error_handler
    @set_event(message=f"Returned to service", classification="Alarm", priority=2, criticity=3)
    def return_to_service(self, user:User=None):
        r"""
        Documentation here
        """
        self.__return_to_service()

        return self, f"{self.tag.get_name()}"

    @logging_error_handler
    def attach(self, machine, tag:Tag):
        
        def attach_observer(machine, tag:Tag):

            observer = MachineObserver(machine)
            query = dict()
            query["action"] = "attach_observer"
            query["parameters"] = {
                "name": tag.name,
                "observer": observer,
            }
            self.tag_engine.request(query)
            self.tag_engine.response()

        attach_observer(machine, tag)

    @set_event(message=f"Updated", classification="Alarm", priority=2, criticity=3)
    def put(
            self, 
            user:User=None,
            name:str=None,
            tag:str=None,
            description:str=None,
            alarm_type:TriggerType=None,
            trigger_value:float=None):
        r"""
        Update alarm configuration

        **Parameters**
        
        * **name** (str): Alarm name
        * **tag** (str): Tag binded to alarm
        * **description** (str): Alarm description
        * **type** (str): Alarm type ['HIGH-HIGH', 'HIGH', 'LOW', 'LOW-LOW', 'BOOL']
        * **trigger_value** (float): Alarm trigger value

        """
        message = ""
        if alarm_type:

            if alarm_type.value.upper() in ["HIGH-HIGH", "HIGH", "LOW", "LOW-LOW", "BOOL"]:

                self.alarm_setpoint.type = alarm_type

                message += f" alarm_type: {alarm_type.value}"

        if trigger_value:
            
            self.alarm_setpoint.value = float(trigger_value)
            message += f" trigger value: {trigger_value}"
        
        if name:

            self._name = name
            message += f" name: {name}"

        if tag:

            self._tag = tag
            message += f" tag: {tag}"

        if description:

            self._description = description
            message += f" description: {description}"
        
        return self, message

    def _get_active_transitions(self):
        r"""
        Gets allowed transitions based on the current state

        **Returns**

        * **(list)**
        """
        result = list()

        current_state = self.current_state
        transitions = self.transitions

        for transition in transitions:

            if transition.source == current_state:

                result.append(transition)

        return result
    
    @logging_error_handler
    def __transition(self, transition_name:str):

        allowed_transitions = self._get_active_transitions()
        for _transition in allowed_transitions:
            
            if f"{_transition.source.name}_to_{_transition.target.name}"==transition_name:
                
                self.send(transition_name)

    @logging_error_handler
    def __return_to_service(self):

        current_state = self.current_state.name.lower()

        if self.state.alarm_status.lower()=="active":

            transition_name = f'{current_state}_to_unack_alarm'

        else:

            transition_name = f'{current_state}_to_normal'
        
        self.__transition(transition_name=transition_name)

    @logging_error_handler
    def get_operator_actions(self)->list:
        
        current_state = self.current_state.name.lower()
            
        if current_state in ("unack_alarm", "rtn_unack"):

            result = {
                "Acknowledge": "acknowledge",
                "Shelve": "shelve",
                "Designed Suppression": "designed_suppression",
                "Remove From Service": "remove_from_service"
            }

        elif current_state=="suppressed_by_design":

            result = {
                "Designed Unsuppression": "designed_unsuppression"
            }

        elif current_state=="out_of_service":

            result = {
                "Return To Service": "return_to_service"
            }

        else:

            result = {
                "Shelve": "shelve",
                "Designed Suppression": "designed_suppression",
                "Remove From Service": "remove_from_service"
            }

        return result

    def serialize(self):
        r"""
        Allows you to serialize alarm to a dict jsonable

        **Return**

        * **alarm_info**: (dict) A jsonable dictionary
        """
        timestamp = self.timestamp
        if timestamp:

            timestamp = timestamp.strftime(self.tag_engine.DATETIME_FORMAT)

        ack_timestamp = self.ack_timestamp
        if ack_timestamp:

            ack_timestamp = ack_timestamp.strftime(self.tag_engine.DATETIME_FORMAT)

        return {
            "identifier": self.identifier,
            "segment": self.segment,
            "manufacturer": self.manufacturer,
            "timestamp": timestamp,
            "name": self.name,
            "tag": self.tag.name,
            "state": self.state.serialize(),
            "alarm_setpoint": self.alarm_setpoint.serialize(),
            "ack_timestamp": ack_timestamp,
            "description": self.description,
            "actions": self.get_operator_actions()
        }





import dash
import dash_bootstrap_components as dbc
from automation.models import StringType


class MachinesComponents:

    @classmethod
    def machine_actions(cls):
        r"""
        Documentation here
        """

        return dash.html.Div([
            dash.dcc.Location(id='machine_actions_page', refresh=False),
            dbc.Row([
                dbc.Col([
                    dbc.InputGroup(
                        [
                            dbc.InputGroupText(id="machine_state_input"),
                            dbc.Select(
                                options=[],
                                id="machine_actions_input"
                            ),
                        ],
                        size="md"
                    )
                ],
                width=12,
                className="col-sm-12 col-md-12")
            ])
        ])
    
    @classmethod
    def machine_attributes(cls, app, machine_name:str):
        r"""
        Documentation here
        """
        machine = app.get_machine(name=StringType(machine_name))
        machine_interval = machine.get_interval()
        threshold = machine.threshold
        on_delay = machine.on_delay
        disable = True
        if hasattr(machine, "buffer_size"):
            disable = False
        if "pfm" in machine_name.lower():
            disable = True
        elif "observer" in  machine_name.lower():
            disable = True
        return dash.html.Div([
            dash.dcc.Location(id='machine_attributes_page', refresh=False),
            dbc.Accordion(
                    [
                        dbc.AccordionItem(
                            [
                                dbc.Row([
                                    dbc.Col(
                                        [
                                            dbc.InputGroup(
                                                [
                                                    dbc.Input(placeholder=f"{machine_name} Current threshold {threshold.value}", type="number", step=0.1, min=0.0, max=600000, id="machine_threshold_input", disabled=disable, n_submit=0), 
                                                    dbc.InputGroupText(threshold.unit, id="threshold_input_text")
                                                ], 
                                                size="md",
                                                className="mb-1"
                                            )
                                        ],
                                        width=12,
                                        className="col-sm-12 col-md-12 col-lg-12"
                                    ),
                                    dbc.Col(
                                        [
                                            dbc.InputGroup(
                                                [
                                                    dbc.Input(placeholder=f"Current Machine Interval {machine_interval}", type="number", step=0.5, min=0.5, max=600000, id="machine_interval_input", disabled=False, n_submit=0), 
                                                    dbc.InputGroupText('s')
                                                ], 
                                                size="md",
                                                className="mb-1"
                                            )
                                        ],
                                        width=12,
                                        className="col-sm-12 col-md-12 col-lg-12"
                                    ),
                                    dbc.Col(
                                        [
                                            dbc.InputGroup(
                                                [
                                                    dbc.Input(placeholder="Buffer Size", type="number", step=1, min=2, max=600000, id="buffer_size_input", disabled=disable, n_submit=0), 
                                                ], 
                                                size="md",
                                                className="mb-1"
                                            )
                                        ],
                                        width=12,
                                        className="col-sm-12 col-md-12 col-lg-12"
                                    ),
                                    dbc.Col(
                                        [
                                            dbc.InputGroup(
                                                [
                                                    dbc.Input(placeholder=f"Current On Delay: {on_delay.value}", type="number", step=1, min=1, max=600000, id="on_delay_input", disabled=False, n_submit=0), 
                                                    dbc.InputGroupText('s')
                                                ], 
                                                size="md",
                                                className="mb-1"
                                            )
                                        ],
                                        width=12,
                                        className="col-sm-12 col-md-12 col-lg-12"
                                    ),
                                ])
                            ],
                            title="Machine Attributes",
                            className="my-3"
                        )
                    ],
                    start_collapsed=False,
                )
            ])

    @classmethod
    def subscription_tag_form(cls, app, machine_name:str):
        r"""
        Documentation here
        """
        machine = app.get_machine(name=StringType(machine_name))
        internal_variables = machine.get_read_only_process_type_variables()
        subscribed_tags_machine = [{"label": "", "value": ""}]
        not_subscribed = [{"label": "", "value": ""}]
        available_tags = [{"label": "", "value": ""}]
        tags = app.cvt._cvt.get_field_tags_names()
        for _tag_name, value in internal_variables.items():

            if value.tag:
                subscribed_tags_machine.append({
                    "label": f"{value.tag.name}->{_tag_name}", "value": value.tag.name
                    })
                
                if value.tag.name in tags:

                    tags.remove(value.tag.name)
                
            else:

                not_subscribed.append({
                    "label": _tag_name, "value": _tag_name
                })

        for tag in tags:

            available_tags.append({
                "label": tag, "value": tag
            })

        return dash.html.Div(
            [
                dash.dcc.Location(id='tags_machine_subscription_page', refresh=False),
                dbc.Accordion(
                    [
                        dbc.AccordionItem(
                            [
                                dbc.Row(
                                    [
                                        dbc.Col(
                                            [
                                                dbc.InputGroup(
                                                    [
                                                        dbc.InputGroupText("Subscribed"),
                                                        dbc.Select(
                                                            options=subscribed_tags_machine,
                                                            id="subscribed_tag_machine_input"
                                                        ),
                                                    ],
                                                    size="md"
                                                )
                                            ],
                                            width=12,
                                            className="col-sm-12 col-md-12"
                                        )
                                    ],
                                    className="form g-3"
                                ),
                                dbc.Row([
                                    dbc.Col(
                                            [
                                                dbc.InputGroup(
                                                    [
                                                        dbc.InputGroupText("Field"), dbc.Select(options=available_tags, id="field_tag_input" )
                                                    ],
                                                    size="md",
                                                    className="mb-6"
                                                )
                                            
                                            ],
                                            width=12,
                                            className="col-sm-12 col-md-6"
                                        ),
                                        
                                        dbc.Col(
                                            [
                                                dbc.InputGroup(
                                                    [
                                                        dbc.InputGroupText("Tag"), 
                                                        dbc.Select(options=not_subscribed, id="internal_tag_input")
                                                    ],
                                                    size="md",
                                                    className="mb-6"
                                                )
                                            ],
                                            width=12,
                                            className="col-sm-12 col-md-6"
                                        ),
                                ]),
                                dbc.Row([
                                    dbc.Col(
                                        dbc.Button(
                                            "Subscribe",
                                            color="primary",
                                            outline=True,
                                            disabled=True,
                                            id="subscribe_tag_machine_button",
                                            className="w-100"
                                        ),
                                        width="auto",
                                        className="d-flex justify-content-center align-items-center"
                                    ),
                                    dbc.Col(
                                        dbc.Button(
                                            "Unsubscribe",
                                            color="danger",
                                            outline=True,
                                            disabled=True,
                                            id="unsubscribe_tag_machine_button",
                                            className="w-100"
                                        ),
                                        width="auto",
                                        className="d-flex justify-content-center align-items-center"
                                    ),
                                ])
                            ],
                            title="Tags - Machine Subscription",
                            className="my-3"
                        )
                    ],
                    start_collapsed=False,
                )
            ]
        )

    @classmethod
    def machines_table(cls)->dash.dash_table.DataTable:
        r"""
        Documentation here
        """
        return dbc.Container(
            dbc.Row(
                dbc.Col(
                    dash.dash_table.DataTable(
                        data=[],
                        columns=[
                            {'name': 'name', 'id': 'name', 'editable': False}, 
                            {'name': 'sampling_time', 'id': 'sampling_time', 'editable': False}, 
                            {'name': 'description', 'id': 'description', 'editable': False}, 
                            {'name': 'state', 'id': 'state', 'editable': False}, 
                            {'name': 'criticity', 'id': 'criticity', 'editable': False}, 
                            {'name': 'priority', 'id': 'priority', 'editable': False},
                            {'name': 'classification', 'id': 'classification', 'editable': False},
                        ],
                        id="machines_datatable",
                        filter_action="native",
                        sort_action="native",
                        sort_mode="multi",
                        selected_columns=[],
                        page_action="native",
                        page_current= 0,
                        page_size= 10,
                        persistence=True,
                        editable=True,
                        persisted_props=['data'],
                        export_format='xlsx',
                        export_headers='display',
                        style_table={'overflowX': 'auto'},
                    ),
                    width=12,
                )
            ),
            fluid=True,
            className="mx-0 px-0"
        )
    
    @classmethod
    def machines_tabs(cls):

        return dash.html.Div(
            [
                dbc.Tabs(
                    [
                        dbc.Tab(label="No Advanced Machine", tab_id="tab-1"),
                    ],
                    id="machines_detailed_tabs",
                    active_tab="tab-1",
                ),
                dash.html.Div(id="machine_detailed_content"),
            ]
        )
    
    @classmethod
    def machine_tab_datatable_content(cls, app, machine_name:str)->dash.dash_table.DataTable:
        r"""
        Documentation here
        """
        def flatten_dict(d, parent_key='', sep='.'): 
            items = [] 
            for k, v in d.items(): 
                
                new_key = f"{parent_key}{sep}{k}" if parent_key else k 
                if isinstance(v, dict): 
                    items.extend(flatten_dict(v, new_key, sep=sep).items()) 
                else: 
                    items.append((new_key, v)) 
                    
            return dict(items)

        machine = app.get_machine(name=StringType(machine_name))
        data = list()
        serialized_data = machine.serialize()
        flattened_data = flatten_dict(serialized_data)
        for key, value in flattened_data.items():

            if isinstance(value, (bool, int, str, float)):

                data.append({"name": key, "value": value})

        return dash.dash_table.DataTable(
                data=data,
                columns=[
                    {'name': 'name', 'id': 'name', 'editable': False}, 
                    {'name': 'value', 'id': 'value', 'editable': False}, 
                ],
                id="machines_detailed_datatable",
                filter_action="native",
                sort_action="native",
                sort_mode="multi",
                row_deletable=False,
                selected_columns=[],
                page_action="native",
                page_size=10,
                persistence=True,
                editable=False,
                persisted_props=['data'],
                export_format='xlsx',
                export_headers='display',
                style_table={'overflowX': 'auto'},
            )
    
    @classmethod
    def machine_tab_content(cls, app, machine_name:str):

        return dash.html.Div(
            [
                dbc.Row([
                    dbc.Col(
                        [
                            dbc.Container(
                            dbc.Row(
                                dbc.Col(
                                    cls.machine_tab_datatable_content(app=app, machine_name=machine_name),
                                    width=12,
                                )
                            ),
                            fluid=True,
                            className="mx-0 px-0"
                        )
                        ],
                        width=6,
                        className="col-sm-12 col-md-6"
                    ),  
                    dbc.Col(
                        [
                            dbc.Row([
                                cls.subscription_tag_form(app=app, machine_name=machine_name)
                            ]),
                            dbc.Row([
                                cls.machine_actions()
                            ]),
                            dbc.Row([
                                cls.machine_attributes(app=app, machine_name=machine_name)
                            ]),
                        ],
                        width=6,
                        className="col-sm-12 col-md-6"
                    )
                ])
            ]
        )


import dash
import dash_bootstrap_components as dbc
from automation.pages.components import Components
from automation.pages.components.gaussian_filter import FilterComponents

dash.register_page(__name__)

layout = dbc.Container(
    [
        dbc.Breadcrumb(
            items=[
                {"label": "Home", "href": "/"},
                {"label": "Tags", "href": "/tags"},
                {"label": "Filter", "active": True},
            ],
        ),
        dash.dcc.Location(id='filter_page', refresh=False),
        dbc.Row(
            [
                dbc.Col(FilterComponents.tags(), width=6, className="col-sm-6 col-md-10"),
                dbc.Col(FilterComponents.last_values(), width=6, className="col-sm-6 col-md-2 ps-0"),
            ],
            className="mb-3"
        ),
        dbc.Row(
            [
                dbc.Col(FilterComponents.current_value_table(), width=12, className="col-sm-12 col-md-12 mb-2"),
                dbc.Col(FilterComponents.plot(), width=12, className="col-sm-12 col-md-12"),
            ],
            className="mb-3"
        ),
        Components.modal_error(
            title="Error",
            modal_id="modal-error-filter-tags",
            button_close_id="close-modal-error-filter",
            body_id="modal-error-filter-tags-body"
        ),
        Components.modal_error(
            title="Success",
            modal_id="modal-success-filter-tags",
            button_close_id="close-modal-success-filter-tags",
            body_id="modal-success--filter-tags-body"
        ),
        Components.modal_confirm(
            title="Confirmation",
            modal_id="modal-update-filter-tags",
            body_id="modal-update-filter-tags-body",
            yes_button_id="update-filter-tags-yes",
            no_button_id="update-filter-tags-no"
        ),
    ],
    fluid=False,
    className="my-3"
)
    


import pytz
from datetime import datetime, timedelta
from flask_restx import Namespace, Resource, fields
from .... import PyAutomation
from ....extensions.api import api
from ....extensions import _api as Api
from ....dbmodels.alarms import AlarmSummary
from .... import _TIMEZONE, TIMEZONE

ns = Namespace('Alarms Summary', description='Alarms Summary')
app = PyAutomation()

alarms_summary_filter_model = api.model("alarms_summary_filter_model",{
    'names': fields.List(fields.String(), required=False),
    'states': fields.List(fields.String(), required=False),
    'tags': fields.List(fields.String(), required=False),
    'greater_than_timestamp': fields.DateTime(required=False, default=datetime.now(pytz.utc).astimezone(TIMEZONE) - timedelta(minutes=30), description=f'Greater than timestamp - DateTime Format: {app.cvt.DATETIME_FORMAT}'),
    'less_than_timestamp': fields.DateTime(required=False, default=datetime.now(pytz.utc).astimezone(TIMEZONE), description=f'Less than timestamp - DateTime Format: {app.cvt.DATETIME_FORMAT}'),
    'timezone': fields.String(required=False, default=_TIMEZONE)
})

    
@ns.route('/filter_by')
class AlarmsSummarygFilterByResource(Resource):

    @api.doc(security='apikey')
    @Api.token_required(auth=True)
    @ns.expect(alarms_summary_filter_model)
    def post(self):
        r"""
        Alarms Summary Filter By
        """
        timezone = _TIMEZONE
        if "timezone" in api.payload:

            timezone = api.payload["timezone"]

        if timezone not in pytz.all_timezones:

            return f"Invalid Timezone", 400

        separator = '.'
        if 'greater_than_timestamp' in api.payload:
            
            greater_than_timestamp = api.payload['greater_than_timestamp']
            api.payload['greater_than_timestamp'] = greater_than_timestamp.replace("T", " ").split(separator, 1)[0] + '.00'
        
        if "less_than_timestamp" in api.payload:

            less_than_timestamp = api.payload['less_than_timestamp']
            api.payload['less_than_timestamp'] = less_than_timestamp.replace("T", " ").split(separator, 1)[0] + '.00'
                
        return app.filter_alarms_by(**api.payload)
    

@ns.route('/lasts/<lasts>')
class LastsAlarmsResource(Resource):

    @api.doc(security='apikey')
    @Api.token_required(auth=True)
    def get(self, lasts:int=10):
        r"""
        Get lasts alarms
        """

        return app.get_lasts_alarms(lasts=int(lasts))
    

@ns.route('/<id>/comments')
class AlarmsSummaryCommentsResource(Resource):

    @api.doc(security='apikey')
    @Api.token_required(auth=True)
    def get(self, id:int):
        r"""
        Get Alarm Summary Comments
        """
        comments = AlarmSummary.get_alarm_summary_comments(id=id)
        return comments, 200



class Buffer(list):
    r"""
    Documentation here
    """

    def __init__(self, size:int=10, roll:str='forward'):
        r"""
        Documentation here
        """
        self._roll_type_allowed = ['forward', 'backward']
        self._size = size
        self.roll = roll

    @property
    def size(self):
        r"""
        Documentation here
        """
        return self._size

    @size.setter
    def size(self, value:int):
        r"""
        Documentation here
        """
        if not isinstance(value, int):

            raise TypeError("Only integers are allowed")

        if value <= 1:

            raise ValueError(f"{value} must be greater than one (1)")
        

        self.__init__(value, roll=self.roll)

    def last(self):
        r"""
        Returns last registered value of the buffer
        """
        if self:
            if self.roll == 'forward':

                return self[-1]
            
            return self[0]
    
    def current(self):
        r"""
        Returns lastest registered value of the buffer
        """  
        if self:
            if self.roll == 'forward':
                
                return self[0]
            
            return self[-1]
        
    def previous_current(self):
        r"""
        Returns lastest registered value of the buffer
        """
        if self:
            if self.roll == 'forward':
                
                return self[1]
            
            return self[-2]

    @property
    def roll(self):
        r"""
        Documentation here
        """
        return self.roll_type

    @roll.setter
    def roll(self, value:str):
        r"""
        Documentation here
        """
        if not isinstance(value, str):

            raise TypeError("Only strings are allowed")

        if value not in self._roll_type_allowed:
            
            raise ValueError(f"{value} is not allowed, you can only use: {self._roll_type_allowed}")

        self.roll_type = value

    def __call__(self, value):
        r"""
        Documentation here
        """
        if self.roll.lower()=='forward':
            
            _len = len(self)
            
            if _len >= self.size:
                
                if _len == self.size:
                    
                    self.pop()
                
                else:

                    for _ in range(_len - self.size):

                        self.pop()

            super(Buffer, self).insert(0, value)

        else:

            _len = len(self)
            
            if _len >= self.size:
                
                if _len == self.size:
                    
                    self.pop(0)
                
                else:

                    for _ in range(_len - self.size):

                        self.pop(0)
                
            super(Buffer, self).append(value)

        return self


from enum import Enum


ACTIONS = {
    "Normal": ['shelve', 'suppress by design', 'out of service', 'disable'],
    "Unacknowledged": ['acknowledge', 'shelve', 'suppress by design', 'out of service', 'silence', 'disable'],
    "Acknowledged": ['shelve', 'suppress by design', 'out of service', 'disable'],
    "RTN Unacknowledged": ['shelve', 'suppress by design', 'out of service', 'disable'],
    "Shelved": ["reset"],
    "Suppressed By Design": ["unsuppress by design"],
    "Out Of Service": ["return to service"]
}


class States(Enum):

    NORM = "Normal"
    UNACK = "Unacknowledged"
    ACKED = "Acknowledged"
    RTNUN = "RTN Unacknowledged"
    SHLVD = "Shelved"
    DSUPR = "Suppressed By Design"
    OOSRV = "Out Of Service"


class Status(Enum):

    ACTV = "Active"
    NACTV = "Not Active"
    ANNCTD = "Annunciated"
    NANNCTD = "Not Annunciated"
    OR = "Not Active or Active"
    SUPR = "Suppressed"
    NA = "Not Applicable"
    NORM = "Normal"
    ABNORM = "Abnormal"


class AlarmAttrs:

    def __init__(
        self, 
        mnemonic: str, 
        state: str, 
        process_condition: str,
        alarm_status: str, 
        annunciate_status: str, 
        acknowledge_status: str,
    ):
        self.__mnemonic = mnemonic
        self.__state = state
        self.__process_condition = process_condition
        self.__alarm_status = alarm_status
        self.__annunciate_status = annunciate_status
        self.__acknowledge_status = acknowledge_status

    @property
    def mnemonic(self):
        r"""
        Documentation here
        """
        return self.__mnemonic

    @property
    def state(self):
        r"""
        Documentation here
        """
        return self.__state
    
    @property
    def process_condition(self):
        r"""
        Documentation here
        """
        return self.__process_condition

    @property
    def alarm_status(self):
        r"""
        Documentation here
        """
        return self.__alarm_status

    @property
    def annunciate_status(self):
        r"""
        Documentation here
        """
        return self.__annunciate_status

    @property
    def acknowledge_status(self):
        r"""
        Documentation here
        """
        return self.__acknowledge_status

    def is_acknowledged(self):
        r"""
        Documentation here
        """

        return self.acknowledge_status == States.ACKED.value

    def serialize(self):
        r"""
        Documentation here
        """
        return {
            'mnemonic': self.mnemonic,
            'state': self.state,
            'process_condition': self.process_condition,
            'alarm_status': self.alarm_status,
            'annunciate_status': self.annunciate_status,
            'acknowledge_status': self.acknowledge_status
        }


class AlarmState:

    NORM = AlarmAttrs(
        mnemonic=States.NORM.name,
        state=States.NORM.value,
        process_condition=Status.NORM.value,
        alarm_status=Status.NACTV.value,
        annunciate_status=Status.NANNCTD.value,
        acknowledge_status=States.ACKED.value
    )
    UNACK = AlarmAttrs(
        mnemonic=States.UNACK.name,
        state=States.UNACK.value,
        process_condition=Status.ABNORM.value,
        alarm_status=Status.ACTV.value,
        annunciate_status=Status.ANNCTD.value,
        acknowledge_status=States.UNACK.value
    )
    ACKED = AlarmAttrs(
        mnemonic=States.ACKED.name,
        state=States.ACKED.value,
        process_condition=Status.ABNORM.value,
        alarm_status=Status.ACTV.value,
        annunciate_status=Status.ANNCTD.value,
        acknowledge_status=States.ACKED.value
    )
    RTNUN = AlarmAttrs(
        mnemonic=States.RTNUN.name,
        state=States.RTNUN.value,
        process_condition=Status.NORM.value,
        alarm_status=Status.NACTV.value,
        annunciate_status=Status.ANNCTD.value,
        acknowledge_status=States.UNACK.value
    )
    SHLVD = AlarmAttrs(
        mnemonic=States.SHLVD.name,
        state=States.SHLVD.value,
        process_condition=Status.NORM.value,
        alarm_status=Status.OR.value,
        annunciate_status=Status.SUPR.value,
        acknowledge_status=Status.NA.value
    )
    DSUPR = AlarmAttrs(
        mnemonic=States.DSUPR.name,
        state=States.DSUPR.value,
        process_condition=Status.NORM.value,
        alarm_status=Status.OR.value,
        annunciate_status=Status.SUPR.value,
        acknowledge_status=Status.NA.value
    )
    OOSRV = AlarmAttrs(
        mnemonic=States.OOSRV.name,
        state=States.OOSRV.value,
        process_condition=Status.NORM.value,
        alarm_status=Status.OR.value,
        annunciate_status=Status.SUPR.value,
        acknowledge_status=Status.NA.value
    )

    _states = [NORM, UNACK, ACKED, RTNUN, SHLVD, DSUPR, OOSRV]

    @classmethod
    def get_state_by_name(cls, state:str):
        r"""
        Documentation here
        """
        _state = States(state)
        for alarm_state in cls._states:

            if _state==alarm_state:

                return alarm_state




from peewee import Proxy, Model

proxy = Proxy()

SQLITE = 'sqlite'
MYSQL = 'mysql'
POSTGRESQL = 'postgresql'


class BaseModel(Model):

    @classmethod
    def read(cls, id:int|str):
        r"""
        Select a single record

        You can use this method to retrieve a single instance matching the given query. 

        This method is a shortcut that calls Model.select() with the given query, but limits the result set to a single row. 
        Additionally, if no model matches the given query, a DoesNotExist exception will be raised.

        **Parameters**

        * **id:** (int), Record ID

        **Returns**

        * **result:** (dict) --> {'message': (str), 'data': (list) row serialized}

        """
        query = cls.select().where(cls.id == id).get_or_none()

        if query:
            
            return query

    @classmethod
    def read_all(cls):
        r"""
        Select all records

        You can use this method to retrieve all instances matching in the database. 

        This method is a shortcut that calls Model.select() with the given query.

        **Parameters**

        **Returns**

        * **result:** (dict) --> {'message': (str), 'data': (list) row serialized}
        """
        data = [query.serialize() for query in cls.select()]

        return data

    @classmethod
    def put(cls, id:str, **fields):
        r""""
        Update a single record

        Once a model instance has a primary key, you UPDATE a field by its id. 
        The model's primary key will not change:
        """     
        if cls.id_exists(id):
            
            query = cls.update(**fields).where(cls.id == id)
            query.execute()
            return query            

    @classmethod
    def id_exists(cls, id:str|int)->bool|None:
        r"""
        Verify if a record exist by its id

        **Parameters**

        * **id:** (int) Record ID

        **Returns**

        * **bool:** If True, so id record exist into database
        """
        return True if cls.get_or_none(id=id) else False

    class Meta:
        database = proxy


from ..utils.units import EngUnit, UnitSerializer, UnitError

class Percentage(EngUnit):
    """Creates a percentage object that can store a percentage value and 
    convert between units of percentage.

    :param value: [int|float] Engineering value\n
    :param unit: [str] Engineering unit\n
    :return: [Length Object]\n

    
    ```
    """
    
    class Units(UnitSerializer):
        percentage = '%' 


    conversions = {

        '%' : 1.0
    }

    def __init__(self, value, unit):

        if unit not in Percentage.Units.list():

            raise UnitError(f"{unit} value is not allowed for {self.__class__.__name__} object - you can use: {Percentage.Units.list()}")
        
        super(Percentage, self).__init__(value=value, unit=unit)


import dash
import dash_mantine_components as dmc
import dash_bootstrap_components as dbc
from dash_iconify import DashIconify
from ...singleton import Singleton


class FileTree:
    r"""
    Documentation here
    """
    def __init__(self):
        
        self.data = None
        self.folder_name = None

    def render(self, data) -> dmc.Accordion:
        r"""
        Documentation here
        """
        self.data = data
        return dmc.Accordion(
            self.build_tree(self.data),
            multiple=True,
            id="opcua_server_tree"
        )

    def flatten(self, l):
        r"""
        Documentation here
        """
        return [item for sublist in l for item in sublist]

    def make_file(self, file_name, key):
        r"""
        Documentation here
        """
        return dmc.Text(
            [
                dash.dcc.Checklist(
                    options=[{'label': '', 'value': f"{self.folder_name}/{key}"}],
                    id={'type': 'file-checklist', 'index': f"{self.folder_name}/{key}"},
                    style={"display": "inline-block"}
                ),
                DashIconify(icon="akar-icons:file"),
                " ",
                file_name
            ],
            style={"paddingTop": '5px'}
        )

    def make_folder(self, folder_name):
        r"""
        Documentation here
        """       
        return [DashIconify(icon="akar-icons:folder"), " ", folder_name]

    def build_tree(self, nodes):
        r"""
        Documentation here
        """
        d = []
        for i, node in enumerate(nodes):
            if node['children']:
                self.folder_name = node['title']
            if node['children']:
                children = self.flatten([self.build_tree(node['children'])])
                d.append(
                    dmc.AccordionItem(
                        children=[
                            dmc.AccordionControl(self.make_folder(node['title'])),
                            dmc.AccordionPanel(children)
                        ],
                        value=f"item-{i}"
                    )
                )
            else:
                d.append(self.make_file(node['title'], node['key']))
        return d


file_tree = FileTree()


class OPCUAComponents(Singleton):

    @classmethod
    def add_server(cls, title:str, modal_id:str, body_id:str, ok_button_id:str, cancel_button_id:str):

        return dash.html.Div(
            [
                dash.dcc.Location(id='communications_page', refresh=False),
                dbc.Modal(
                    [
                        dbc.ModalHeader(dbc.ModalTitle(title), close_button=True),
                        dbc.ModalBody(
                            dash.html.Div(
                                [
                                    dash.html.H6('Server Information'),  # This is the title
                                    dash.html.Div([
                                        dbc.InputGroup([dbc.InputGroupText("Name"), dbc.Input(placeholder="Server 1", id="opcua_client_name_input")], size="sm", className="mb-1"),
                                        dbc.InputGroup([dbc.InputGroupText("Host"), dbc.Input(placeholder="127.0.0.1", id="opcua_client_host_input")], size="sm", className="mb-1"),
                                        dbc.InputGroup([dbc.InputGroupText("Port"), dbc.Input(placeholder=4840, id="opcua_client_port_input", type="number")], size="sm"),
                                    ], style={'border': '1px solid black', 'padding': '10px'}, className="mb-2"),

                                    dash.html.H6('Security Settings'),  # This is the title
                                    dash.html.Div([
                                        dbc.InputGroup(
                                            [
                                                dbc.InputGroupText("Security Policy"),
                                                dbc.Select(
                                                    options=[
                                                        {"label": "None", "value": None},
                                                        {"label": "Basic128Rsa15", "value": "Basic128Rsa15"},
                                                        {"label": "Basic256Sha256", "value": "Basic256Sha256"},
                                                        {"label": "Aes128Sha256RsaOaep", "value": "Aes128Sha256RsaOaep"},
                                                        {"label": "Aes256Sha256RsaPss", "value": "Aes256Sha256RsaPss"},
                                                    ]
                                                ),
                                            ],
                                            size="sm",
                                            className="mb-1"
                                        ),
                                        dbc.InputGroup(
                                            [
                                                dbc.InputGroupText("Message Security Mode"),
                                                dbc.Select(
                                                    options=[
                                                        {"label": "None", "value": None},
                                                        {"label": "Sign", "value": "Sign"},
                                                        {"label": "Sign & Encrypt", "value": "Sign & Encrypt"},
                                                    ]
                                                )
                                            ],
                                            size="sm"
                                        )
                                    ], style={'border': '1px solid black', 'padding': '10px'}, className="mb-2"),

                                    dash.html.H6('Authentication Settings'),  # This is the title
                                    dash.html.Div([
                                        dbc.InputGroup([
                                            dbc.RadioItems(
                                                id="radio-1",
                                                options=[{"label": "Anonymous", "value": 1}],
                                            ),
                                        ], className="mb-3"),
                                        dash.html.Hr(),
                                        dbc.InputGroup([
                                            dbc.Row([
                                                dbc.Col([
                                                    dbc.RadioItems(
                                                        id="radio-2",
                                                        options=[{"label": "", "value": 2}]
                                                    ),
                                                ],
                                                width=1),
                                                dbc.Col([
                                                    dbc.InputGroup([dbc.InputGroupText("Username"), dbc.Input(disabled=True)], size="sm", className="mb-1"),
                                                    dbc.InputGroup([dbc.InputGroupText("Password"), dbc.Input(disabled=True)], size="sm"),
                                                ],
                                                width=11)
                                            ])                                            
                                        ], className="mb-3"),
                                        dash.html.Hr(),
                                        dbc.InputGroup([
                                            dbc.Row([
                                                dbc.Col([
                                                    dbc.RadioItems(
                                                        id="radio-2",
                                                        options=[{"label": "", "value": 2}]
                                                    ),
                                                ],
                                                width=1),
                                                dbc.Col([
                                                    dbc.InputGroup([dbc.InputGroupText("Certificate"), dbc.Input(disabled=True)], size="sm", className="mb-1"),
                                                    dbc.InputGroup([dbc.InputGroupText("Private Key"), dbc.Input(disabled=True)], size="sm"),
                                                ],
                                                width=11)
                                            ])                                            
                                        ], className="mb-3"),
                                    ], style={'border': '1px solid black', 'padding': '10px'}),
                                ]
                            ),
                            id=body_id),
                        dbc.ModalFooter(
                            [
                                dbc.Button(
                                    "OK",
                                    id=ok_button_id,
                                    className="float-start",
                                    n_clicks=0,
                                ),
                                dbc.Button(
                                    "Cancel",
                                    id=cancel_button_id,
                                    className="ms-auto",
                                    n_clicks=0,
                                )
                            ]
                        ),
                    ],
                    id=modal_id,
                    centered=True,
                    is_open=False,
                ),
            ]
        )
    
    @classmethod
    def remove_server(cls, title:str, modal_id:str, body_id:str, ok_button_id:str, cancel_button_id:str):

        return dash.html.Div(
            [
                dash.dcc.Location(id='communications_page', refresh=False),
                dbc.Modal(
                    [
                        dbc.ModalHeader(dbc.ModalTitle(title), close_button=True),
                        dbc.ModalBody(
                            dash.html.Div(
                                [
                                    dash.html.H6('Server Information'),  # This is the title
                                    dash.html.Div([
                                        dbc.InputGroup([dbc.InputGroupText("Client Name"), dbc.Select(
                                                        options=[
                                                            {"label": "postgresql", "value": "postgresql"},
                                                            {"label": "mysql", "value": "mysql"},
                                                            {"label": "sqlite", "value": "sqlite"}
                                                        ],
                                                        id="opcua_client_names_options"
                                                    )], size="sm", className="mb-1")
                                    ], style={'border': '1px solid black', 'padding': '10px'}, className="mb-2"),
                                ]
                            ),
                            id=body_id),
                        dbc.ModalFooter(
                            [
                                dbc.Button(
                                    "OK",
                                    id=ok_button_id,
                                    className="float-start",
                                    n_clicks=0,
                                ),
                                dbc.Button(
                                    "Cancel",
                                    id=cancel_button_id,
                                    className="ms-auto",
                                    n_clicks=0,
                                )
                            ]
                        ),
                    ],
                    id=modal_id,
                    centered=True,
                    is_open=False,
                ),
            ]
        )
    
    @classmethod
    def download_node_info(cls, title:str, modal_id:str, body_id:str, ok_button_id:str, cancel_button_id:str):

        return dash.html.Div(
            [
                dash.dcc.Location(id='communications_page', refresh=False),
                dbc.Modal(
                    [
                        dbc.ModalHeader(dbc.ModalTitle(title), close_button=True),
                        dbc.ModalBody(
                            dash.html.Div(
                                [
                                    dash.html.H6('Server Information'),  # This is the title
                                    dash.html.Div([
                                        dbc.InputGroup([dbc.InputGroupText("Client Name"), dbc.Select(
                                                        options=[],
                                                        id="download_opcua_client_names_options"
                                                    )], size="sm", className="mb-1")
                                    ], style={'border': '1px solid black', 'padding': '10px'}, className="mb-2"),
                                ]
                            ),
                            id=body_id),
                        dbc.ModalFooter(
                            [
                                dbc.Button(
                                    "OK",
                                    id=ok_button_id,
                                    className="float-start",
                                    n_clicks=0,
                                ),
                                dbc.Button(
                                    "Cancel",
                                    id=cancel_button_id,
                                    className="ms-auto",
                                    n_clicks=0,
                                )
                            ]
                        ),
                    ],
                    id=modal_id,
                    centered=True,
                    is_open=False,
                ),
            ]
        )
    
    @classmethod
    def flatten_dict(cls, d, parent_name=''): 
        flattened_data = [] 
        current_name = f"{parent_name}.{d['title']}" if parent_name else d['title'] 
        current_info = { 
            'name': current_name, 
            'namespace': d['key'], 
            'NodeClass': d['NodeClass'] 
        } 
        flattened_data.append(current_info) 
        for child in d.get('children', []): 
            flattened_data.extend(cls.flatten_dict(child, current_name)) 
        return flattened_data
    @classmethod
    def get_opcua_tree(cls, app):

        clients = app.automation.get_opcua_clients()
        data = list()
        for client_name, _ in clients.items():
            
            opcua_tree = app.automation.get_opcua_tree(client_name=client_name)
            opcua_tree = opcua_tree[0]['Objects'][0]
            opcua_tree["title"] = client_name
            data.append(opcua_tree)

        data = file_tree.render(data)

        return data

    def data_access_view_table(self, data:list=[])->dash.dash_table.DataTable:
        r"""
        Documentation here
        """
        self.data = data
        return dbc.Container(
            dbc.Row(
                dbc.Col(
                    dash.dash_table.DataTable(
                        data=self.data,
                        columns=[ 
                            {'name': 'server', 'id': 'server', 'editable': False}, 
                            {'name': 'namespace', 'id': 'namespace', 'editable': False}, 
                            {'name': 'data_type', 'id': 'data_type', 'editable': False}, 
                            {'name': 'display_name', 'id': 'display_name', 'editable': False}, 
                            {'name': 'value', 'id': 'value', 'editable': False},
                            {'name': 'source_timestamp', 'id': 'source_timestamp', 'editable': False},
                            {'name': 'status_code', 'id': 'status_code', 'editable': False}
                        ],
                        id="data_access_view_datatable",
                        # row_selectable='single',
                        page_action="native",
                        page_current= 0,
                        page_size= 10,
                        persistence=True,
                        editable=False,
                        persisted_props=['data'],
                        export_format='xlsx',
                        export_headers='display',
                        style_table={'overflowX': 'auto'},
                    ),
                    width=12,
                )
            ),
            fluid=True,
            className="mx-0 px-0"
        ) 
        
    def update_data_access_view(self, namespace, value, timestamp)->dash.dash_table.DataTable:
        r"""
        Documentation here
        """
        _data = self.data.copy()
        for counter, data in enumerate(_data):

            if namespace==data["namespace"]:

                self.data[counter]["value"] = value
                self.data[counter]["source_timestamp"] = timestamp
        
        return self.data


from enum import Enum

class TriggerType(Enum):

    HH = "HIGH-HIGH"
    H = "HIGH"
    L = "LOW"
    LL = "LOW-LOW"
    B = "BOOL"
    NONE = "NOT DEFINED"


class Trigger:

    def __init__(self):

        self.__type = TriggerType.NONE
        self.__value = None

    @property
    def type(self):
        r"""
        Documentation here
        """

        return self.__type

    @type.setter
    def type(self, _type:str):

        self.__type = TriggerType(_type)

    @property
    def value(self):
        r"""
        Documentation here
        """
        return self.__value

    @value.setter
    def value(self, value):

        if self.type==TriggerType.B:

            if isinstance(value, bool):

                self.__value = value

            else:

                self.__value = bool(value)

        else:

            if isinstance(value, (float, int)):

                self.__value = value

    def serialize(self):

        return {
            "type": self.type.value,
            "value": self.value
        }


import threading, logging
from ..singleton import Singleton
from ..dbmodels import (
    Variables, 
    Units,
    DataTypes
    )
from ..variables import VARIABLES, DATATYPES

class BaseLogger(Singleton):

    def __init__(self):

        self._db = None
        self.is_history_logged = True

    def set_db(self, db):
        r"""Documentation here
        """
        self._db = db

    def get_db(self):
        r"""
        Documentation here
        """
        return self._db
    
    def check_connectivity(self):

        try: 
            
            if self._db:
                
                self._db.execute_sql('SELECT 1;')
                
                return True
            
            return False
        
        except: 
            
            return False
    
    def set_is_history_logged(self, value:bool=False):
        r"""Documentation here
        """
        self.is_history_logged = value
    
    def stop_db(self):
        r""""
        Documentation here
        """
        try:
            if self._db:
                self._db.close()
                self._db = None
        except:

            pass

    def create_tables(self, tables):
        r"""
        Documentation here
        """
        if not self._db:
            
            return
        
        self._db.create_tables(tables, safe=True)
        self.__init_default_variables_schema()
        self.__init_default_datatypes_schema()
        self.__init_default_roles_schema()

    def __init_default_roles_schema(self):
        r"""
        Documentatio here
        """
        from ..dbmodels import Roles
        for role in Roles.__defaults__:

            if not Roles.name_exist(name=role['name']):

                Roles.create(**role)

    def __init_default_variables_schema(self):
        r"""
        Documentation here
        """
        for variable, units in VARIABLES.items():
    
            if not Variables.name_exist(variable):
                
                Variables.create(name=variable)

            for name, unit in units.items():

                if not Units.name_exist(unit):

                    Units.create(name=name, unit=unit, variable=variable)

    def __init_default_datatypes_schema(self):
        r"""
        Documentation here
        """
        for datatype in DATATYPES:

            DataTypes.create(name=datatype["value"])

    def drop_tables(self, tables):
        r"""
        Documentation here
        """
        if not self._db:
            
            return

        self._db.drop_tables(tables, safe=True)

class BaseEngine(Singleton):
    r"""
    Alarms logger Engine class for Tag thread-safe database logging.
    """
    logger = BaseLogger()

    def __init__(self):

        super(BaseEngine, self).__init__()
        self._request_lock = threading.Lock()
        self._response_lock = threading.Lock()
        self._response = None
        self._response_lock.acquire()

    def set_db(self, db):
        r"""
        Sets the database, it supports SQLite and Postgres,
        in case of SQLite, the filename must be provided.

        if app mode is "Development" you must use SQLite Databse

        **Parameters:**

        * **dbfile** (str): a path to database file.
        * *drop_table** (bool): If you want to drop table.
        * **cascade** (bool): if there are some table dependency, drop it as well
        * **kwargs**: Same attributes to a postgres connection.

        **Returns:** `None`

        Usage:

        ```python
        >>> app.set_db(dbfile="app.db")
        ```
        """
        self.logger.set_db(db)

    def stop_db(self):
        r"""
        Documentation here
        """
        self.logger.stop_db()

    def get_db(self):
        r"""
        Returns a DB object
        """
        return self.logger.get_db()

    def query(self, query:dict)->dict:
        r"""
        Documentation here
        """
        self.request(query)
        result = self.response()
        if result["result"]:
            return result["response"]

    def request(self, query:dict):
        r"""Documentation here

        # Parameters

        - 

        # Returns

        - 
        """
        self._request_lock.acquire()
        action = query["action"]
        error_msg = f"Error in BaseEngine: {action}"

        try:

            if hasattr(self.logger, action):

                method = getattr(self.logger, action)
                
                if 'parameters' in query:
                    
                    resp = method(**query["parameters"])
                
                else:

                    resp = method()

            self.__true_response(resp)

        except Exception as e:

            self.__log_error(e, error_msg)

        self._response_lock.release()

    def __log_error(self, e:Exception, msg:str):
        r"""
        Documentation here
        """
        logger = logging.getLogger("pyautomation")
        logger.error(f"{e} Message: {msg}")
        self._response = {
            "result": False,
            "response": None
        }

    def response(self):
        r"""Documentation here

        # Parameters

        - 

        # Returns

        - 
        """
        self._response_lock.acquire()
        result = self._response
        self._request_lock.release()
        return result
    
    def __true_response(self, resp):
        r"""
        Documentation here
        """
        self._response = {
            "result": True,
            "response": resp
        }

    def __getstate__(self):

        self._response_lock.release()
        state = self.__dict__.copy()
        del state['_request_lock']
        del state['_response_lock']
        return state

    def __setstate__(self, state):
        
        self.__dict__.update(state)
        self._request_lock = threading.Lock()
        self._response_lock = threading.Lock()

        self._response_lock.acquire()



# -*- coding: utf-8 -*-
"""pyhades/managers/logger.py

This module implements Logger Manager.
"""
import logging, queue
from ..singleton import Singleton
from ..logger.datalogger import DataLoggerEngine
from ..logger.logdict import  LogTable
from ..logger.alarms import AlarmsLoggerEngine
from ..logger.events import EventsLoggerEngine
from ..logger.users import UsersLoggerEngine
from ..logger.logs import LogsLoggerEngine
from ..logger.machines import MachinesLoggerEngine
from ..logger.opcua_server import OPCUAServerLoggerEngine
from ..tags import CVTEngine, TagObserver
from ..modules.users.users import User
from ..utils.decorators import logging_error_handler
from ..dbmodels import (
    Manufacturer,
    Segment,
    Tags, 
    TagValue, 
    AlarmTypes,
    AlarmStates, 
    Alarms,  
    AlarmSummary, 
    Variables, 
    Units, 
    DataTypes,
    OPCUA,
    Users,
    Roles,
    Events,
    Logs,
    Machines,
    TagsMachines,
    AccessType,
    OPCUAServer,
    BaseModel
)


class DBManager(Singleton):
    r"""
    Database Manager class for database logging settings.
    """

    def __init__(self, period:float=1.0, delay:float=1.0, drop_tables:bool=False):

        self._period = period
        self._delay = delay
        self._drop_tables = drop_tables
        self._tag_queue = queue.Queue()
        self.engine = CVTEngine()
        self._logging_tags = LogTable()
        self._logger = DataLoggerEngine()
        self.alarms_logger = AlarmsLoggerEngine()
        self.events_logger = EventsLoggerEngine()
        self.users_logger = UsersLoggerEngine()
        self.logs_logger = LogsLoggerEngine()
        self.machines_logger = MachinesLoggerEngine()
        self.opcuaserver_logger = OPCUAServerLoggerEngine()
        self._tables = [
            Manufacturer,
            Segment,
            Variables, 
            Units, 
            DataTypes, 
            Tags, 
            TagValue, 
            AlarmTypes,
            AlarmStates,
            Alarms,
            AlarmSummary,
            OPCUA,
            Roles,
            Users,
            Events,
            Logs,
            Machines,
            TagsMachines,
            AccessType,
            OPCUAServer
        ]

        self._extra_tables = []
        
    def get_queue(self)->queue.Queue:
        r"""
        Documentation here
        """
        return self._tag_queue

    def set_db(self, db, is_history_logged:bool=False):
        r"""
        Initialize a new DB Object SQLite - Postgres - MySQL

        **Parameters**

        * **db** (db object): Sqlite - Postgres or MySql db object

        **Returns** `None`
        """
        self._logger.set_db(db)
        self._logger.logger.set_is_history_logged(value=is_history_logged)
        self.alarms_logger.set_db(db)
        self.alarms_logger.logger.set_is_history_logged(value=is_history_logged)
        self.events_logger.set_db(db)
        self.events_logger.logger.set_is_history_logged(value=is_history_logged)
        self.users_logger.set_db(db)
        self.logs_logger.set_db(db)
        self.logs_logger.logger.set_is_history_logged(value=is_history_logged)
        self.machines_logger.set_db(db)
        self.opcuaserver_logger.logger.set_db(db)
        
    def get_db(self):
        r"""
        Returns a DB object
        """
        return self._logger.get_db()

    def set_dropped(self, drop_tables:bool):
        r"""
        Allows to you set a flag variable to drop database tables when run app.

        **Parameters**

        * **drop_tables** (bool) If True, drop all tables define in the app an initialized it in blank.

        **Returns**

        * **None**
        """
        self._drop_tables = drop_tables

    def get_dropped(self)->bool:
        r"""
        Gets flag variables to drop tables when initialize the app

        **Return**

        * **drop_tables** (bool) Flag variables to drop table
        """
        return self._drop_tables

    def register_table(self, cls:BaseModel):
        r"""
        Allows to you register a new database model

        **Parameters**

        * **cls* (BaseModel): A class that inherit from BaseModel

        """
        self._tables.append(cls)

    def get_db_table(self, tablename:str):
        r"""
        Documentation here
        """
        for table in self._tables:

            if table._meta.table_name.lower()==tablename.lower():

                return table
            
        return None

    def create_tables(self):
        r"""
        Creates default tables and tables registered with method *register_table*
        """
        self._tables.extend(self._extra_tables)
        self._logger.create_tables(self._tables)
        self.alarms_logger.create_tables(self._tables)

    def drop_tables(self):
        r"""
        Drop all tables defined
        """
        tables = self._tables
        
        self._logger.drop_tables(tables)

    def clear_default_tables(self):
        r"""
        If you want initialize any PyHades app without default tables, you can use this method
        """
        self._tables = []

    def get_tags(self)->dict:
        r"""
        Gets all tag defined in tag's repository
        """
        return self._logger.get_tags()
    
    def get_alarms(self)->dict:
        r"""
        Gets all tag defined in tag's repository
        """

        return self.alarms_logger.get_alarms()

    @logging_error_handler
    def set_tag(
        self, 
        tag:str, 
        unit:str, 
        data_type:str, 
        description:str,
        display_name:str="", 
        min_value:float=None, 
        max_value:float=None, 
        tcp_source_address:str=None, 
        node_namespace:str=None):
        r"""
        Sets tag to Database

        **Parameters**

        * **tag** (str):
        * **unit** (str):
        * **data_type** (str):
        * **description** (str):
        * **min_value** (float)[Optional]:
        * **max_value** (float)[Optional]:
        * **tcp_source_address** (str)[Optional]:
        * **node_namespace** (str)[Optional]:
        """
        self._logger.set_tag(
            tag=tag,  
            unit=unit,
            data_type=data_type,
            description=description,
            display_name=display_name,
            min_value=min_value,
            max_value=max_value,
            tcp_source_address=tcp_source_address,
            node_namespace=node_namespace
        )

    def set_tags(self):
        r"""
        Allows to you define all tags added with *add_tag* method
        """
        for period in self._logging_tags.get_groups():
            
            tags = self._logging_tags.get_tags(period)
        
            for tag, unit, data_type, description, display_name, min_value, max_value, tcp_source_address, node_namespace in tags:

                self.set_tag(
                    tag=tag,
                    unit=unit, 
                    data_type=data_type, 
                    description=description, 
                    display_name=display_name,
                    min_value=min_value, 
                    max_value=max_value, 
                    tcp_source_address=tcp_source_address, 
                    node_namespace=node_namespace)

    def init_database(self):
        r"""
        Initializes all databases.
        """
        if self.get_dropped():
            try:
                self.drop_tables()
            except Exception as e:
                error = str(e)
                logger = logging.getLogger("pyautomation")
                logger.error("Database:{}".format(error))
        
        self.create_tables()

    def stop_database(self):
        r"""
        Documentation here
        """
        self._logger.stop_db()

    def get_opcua_clients(self):
        r"""
        Documentation here
        """
        return OPCUA.read_all()

    # USERS METHODS
    def set_role(self, name:str, level:int, identifier:str):
        r"""
        Documentation here
        """
        return self.users_logger.set_role(name=name, level=level, identifier=identifier)

    def set_user(self, user:User):
        r"""
        Documentation here
        """
        return self.users_logger.set_user(user=user)
    
    def login(self, password:str, username:str="", email:str=""):
        r"""
        Documentation here
        """
        return self.users_logger.login(password=password, username=username, email=email)

    def summary(self)->dict:
        r"""
        Get database manager summary

        **Returns**

        * **summary** (dict): Database summary
        """
        result = dict()

        result["period"] = self.get_period()
        result["tags"] = self.get_tags()
        result["delay"] = self.get_delay()

        return result
    
    def attach(self, tag_name:str):

        observer = TagObserver(self._tag_queue)
        self.engine.attach(name=tag_name, observer=observer)



from ..utils.decorators import decorator


@decorator
def filter(func, args, kwargs):
    r"""
    Documentation here
    """
    cvt = args[0]
    tag_id = kwargs["id"]
    value = kwargs["value"]
    tag = cvt.get_tag(id=tag_id)
    if tag.gaussian_filter:
        
        kwargs["value"] = tag.filter(value, threshold=tag.gaussian_filter_threshold, r_value=tag.gaussian_filter_r_value)

        return func(*args, **kwargs)
    
    return func(*args, **kwargs)
